# test_bayes_net.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>

'''
These are unit tests for the Markov network model of pomegranate.
'''

from __future__ import division

from pomegranate import JointProbabilityTable
from pomegranate import MarkovNetwork
from pomegranate.io import DataGenerator
from pomegranate.io import DataFrameGenerator

from nose.tools import with_setup
from nose.tools import assert_equal
from nose.tools import assert_raises
from nose.tools import assert_true
from nose.tools import assert_almost_equal

from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

import pandas
import random, numpy
import sys

def setup_markov_network_int():
	global d1, d2, d3
	global model1, model2, model3, model4

	d1 = JointProbabilityTable([
		[0, 0, 0.1],
		[0, 1, 0.2],
		[1, 0, 0.4],
		[1, 1, 0.3]], [0, 1])

	d2 = JointProbabilityTable([
		[0, 0, 0, 0.05],
		[0, 0, 1, 0.15],
		[0, 1, 0, 0.07],
		[0, 1, 1, 0.03],
		[1, 0, 0, 0.12],
		[1, 0, 1, 0.18],
		[1, 1, 0, 0.10],
		[1, 1, 1, 0.30]], [1, 2, 3])

	d3 = JointProbabilityTable([
		[0, 0, 0, 0.08],
		[0, 0, 1, 0.12],
		[0, 1, 0, 0.11],
		[0, 1, 1, 0.19],
		[1, 0, 0, 0.04],
		[1, 0, 1, 0.06],
		[1, 1, 0, 0.23],
		[1, 1, 1, 0.17]], [2, 3, 4])

	model1 = MarkovNetwork([d1])
	model1.bake()

	model2 = MarkovNetwork([d1, d2])
	model2.bake()

	model3 = MarkovNetwork([d1, d2, d3])
	model3.bake()

	model4 = MarkovNetwork([d1, d3])
	model4.bake()

def setup_markov_network_str():
	global d1, d2, d3
	global model1, model2, model3, model4

	d1 = JointProbabilityTable([
		['0', '0', 0.1],
		['0', '1', 0.2],
		['1', '0', 0.4],
		['1', '1', 0.3]], [0, 1])

	d2 = JointProbabilityTable([
		['0', '0', '0', 0.05],
		['0', '0', '1', 0.15],
		['0', '1', '0', 0.07],
		['0', '1', '1', 0.03],
		['1', '0', '0', 0.12],
		['1', '0', '1', 0.18],
		['1', '1', '0', 0.10],
		['1', '1', '1', 0.30]], [1, 2, 3])

	d3 = JointProbabilityTable([
		['0', '0', '0', 0.08],
		['0', '0', '1', 0.12],
		['0', '1', '0', 0.11],
		['0', '1', '1', 0.19],
		['1', '0', '0', 0.04],
		['1', '0', '1', 0.06],
		['1', '1', '0', 0.23],
		['1', '1', '1', 0.17]], [2, 3, 4])

	model1 = MarkovNetwork([d1])
	model1.bake()

	model2 = MarkovNetwork([d1, d2])
	model2.bake()

	model3 = MarkovNetwork([d1, d2, d3])
	model3.bake()

	model4 = MarkovNetwork([d1, d3])
	model4.bake()

def setup_markov_network_bool():
	global d1, d2, d3
	global model1, model2, model3, model4

	d1 = JointProbabilityTable([
		[False, False, 0.1],
		[False, True,  0.2],
		[True,  False, 0.4],
		[True,  True,  0.3]], [0, 1])

	d2 = JointProbabilityTable([
		[False, False, False, 0.05],
		[False, False, True,  0.15],
		[False, True,  False, 0.07],
		[False, True,  True,  0.03],
		[True,  False, False, 0.12],
		[True,  False, True,  0.18],
		[True,  True,  False, 0.10],
		[True,  True,  True,  0.30]], [1, 2, 3])

	d3 = JointProbabilityTable([
		[False, False, False, 0.08],
		[False, False, True,  0.12],
		[False, True,  False, 0.11],
		[False, True,  True,  0.19],
		[True,  False, False, 0.04],
		[True,  False, True,  0.06],
		[True,  True,  False, 0.23],
		[True,  True,  True,  0.17]], [2, 3, 4])

	model1 = MarkovNetwork([d1])
	model1.bake()

	model2 = MarkovNetwork([d1, d2])
	model2.bake()

	model3 = MarkovNetwork([d1, d2, d3])
	model3.bake()

	model4 = MarkovNetwork([d1, d3])
	model4.bake()

def setup_markov_network_mixed():
	global d1, d2, d3
	global model1, model2, model3, model4

	d1 = JointProbabilityTable([
		[False, 'blue', 0.1],
		[False, 'red',  0.2],
		[True,  'blue', 0.4],
		[True,  'red',  0.3]], [0, 1])

	d2 = JointProbabilityTable([
		['blue', False, 0, 0.05],
		['blue', False, 1, 0.15],
		['blue', True,  0, 0.07],
		['blue', True,  1, 0.03],
		['red',  False, 0, 0.12],
		['red',  False, 1, 0.18],
		['red',  True,  0, 0.10],
		['red',  True,  1, 0.30]], [1, 2, 3])

	d3 = JointProbabilityTable([
		[False, 0, 'a', 0.08],
		[False, 0, 'b', 0.12],
		[False, 1, 'a', 0.11],
		[False, 1, 'b', 0.19],
		[True,  0, 'a', 0.04],
		[True,  0, 'b', 0.06],
		[True,  1, 'a', 0.23],
		[True,  1, 'b', 0.17]], [2, 3, 4])

	model1 = MarkovNetwork([d1])
	model1.bake()

	model2 = MarkovNetwork([d1, d2])
	model2.bake()

	model3 = MarkovNetwork([d1, d2, d3])
	model3.bake()

	model4 = MarkovNetwork([d1, d3])
	model4.bake()

def teardown():
	pass

def test_initialize():
	assert_raises(ValueError, MarkovNetwork, [])

	d1 = JointProbabilityTable([
		[0, 0, 0.2],
		[0, 1, 0.2],
		[1, 0, 0.4],
		[1, 1, 0.2]], [0, 1])

	model = MarkovNetwork([d1])

@with_setup(setup_markov_network_int, teardown)
def test_structure():
	assert_equal(model1.structure, ((0, 1),))
	assert_equal(model2.structure, ((0, 1), (1, 2, 3)))
	assert_equal(model3.structure, ((0, 1), (1, 2, 3), (2, 3, 4)))
	assert_equal(model4.structure, ((0, 1), (2, 3, 4)))

@with_setup(setup_markov_network_int, teardown)
def test_partition():
	model3.bake()
	assert_true(model3.partition != float("inf"))

	model3.bake(calculate_partition=False)
	assert_true(model3.partition == float("inf"))

@with_setup(setup_markov_network_int, teardown)
def test_d():
	assert_equal(model1.d, 2)
	assert_equal(model2.d, 4)
	assert_equal(model3.d, 5)
	assert_equal(model4.d, 5)

@with_setup(setup_markov_network_mixed, teardown)
def test_d_mixed():
	assert_equal(model1.d, 2)
	assert_equal(model2.d, 4)
	assert_equal(model3.d, 5)
	assert_equal(model4.d, 5)

@with_setup(setup_markov_network_int, teardown)
def test_log_probability_int():
	x = [1, 0]
	logp1 = model1.log_probability(x)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = [1, 0, 1, 1]
	logp1 = model2.log_probability(x)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -3.7297014467295373)
	
	x = [1, 0, 1, 0, 1]
	logp1 = model3.log_probability(x)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -4.429966143312331)

	logp3 = model4.log_probability(x)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, -3.7297014486341915)
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_str, teardown)
def test_log_probability_str():
	x = ['1', '0']
	logp1 = model1.log_probability(x)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = ['1', '0', '1', '1']
	logp1 = model2.log_probability(x)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -3.7297014467295373)
	
	x = ['1', '0', '1', '0', '1']
	logp1 = model3.log_probability(x)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -4.429966143312331)

	logp3 = model4.log_probability(x)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, -3.7297014486341915)
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_bool, teardown)
def test_log_probability_bool():
	x = [True, False]
	logp1 = model1.log_probability(x)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = [True, False, True, True]
	logp1 = model2.log_probability(x)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -3.7297014467295373)
	
	x = [True, False, True, False, True]
	logp1 = model3.log_probability(x)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -4.429966143312331)

	logp3 = model4.log_probability(x)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, -3.7297014486341915)
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_mixed, teardown)
def test_log_probability_mixed():
	x = [True, 'blue']

	logp1 = model1.log_probability(x)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = [True, 'blue', True, 1]
	logp1 = model2.log_probability(x)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -3.7297014467295373)
	
	x = [1, 'blue', True, 0, 'b']
	logp1 = model3.log_probability(x)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_raises(AssertionError, assert_almost_equal, logp1, logp2)
	assert_almost_equal(logp1, -4.429966143312331)

	logp3 = model4.log_probability(x)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, -3.7297014486341915)
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_int, teardown)
def test_log_probability_unnormalized_int():
	x = [1, 0]
	logp1 = model1.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = [1, 0, 1, 1]
	logp1 = model2.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.03))
	
	x = [1, 0, 1, 0, 1]
	logp1 = model3.log_probability(x, unnormalized=True)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.07 * 0.06))

	logp3 = model4.log_probability(x, unnormalized=True)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, numpy.log(0.4 * 0.06))
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_str, teardown)
def test_log_probability_unnormalized_str():
	x = ['1', '0']
	logp1 = model1.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = ['1', '0', '1', '1']
	logp1 = model2.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.03))
	
	x = ['1', '0', '1', '0', '1']
	logp1 = model3.log_probability(x, unnormalized=True)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.07 * 0.06))

	logp3 = model4.log_probability(x, unnormalized=True)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, numpy.log(0.4 * 0.06))
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_bool, teardown)
def test_log_probability_unnormalized_bool():
	x = [True, False]
	logp1 = model1.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = [True, False, True, True]
	logp1 = model2.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.03))
	
	x = [True, False, True, False, True]
	logp1 = model3.log_probability(x, unnormalized=True)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.07 * 0.06))

	logp3 = model4.log_probability(x, unnormalized=True)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, numpy.log(0.4 * 0.06))
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_mixed, teardown)
def test_log_probability_unnormalized_mixed():
	x = [True, 'blue']
	logp1 = model1.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x)

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4))

	x = [True, 'blue', True, 1]
	logp1 = model2.log_probability(x, unnormalized=True)
	logp2 = d1.log_probability(x[:2]) + d2.log_probability(x[1:])

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.03))
	
	x = [1, 'blue', True, 0, 'b']
	logp1 = model3.log_probability(x, unnormalized=True)
	logp2 = (d1.log_probability(x[:2]) + d2.log_probability(x[1:4])
		+ d3.log_probability(x[2:]))

	assert_almost_equal(logp1, logp2)
	assert_almost_equal(logp1, numpy.log(0.4 * 0.07 * 0.06))

	logp3 = model4.log_probability(x, unnormalized=True)
	logp4 = d1.log_probability(x[:2]) + d3.log_probability(x[2:])

	assert_almost_equal(logp3, logp4)
	assert_almost_equal(logp3, numpy.log(0.4 * 0.06))
	assert_raises(AssertionError, assert_almost_equal, logp1, logp3)

@with_setup(setup_markov_network_int, teardown)
def test_predict_int():
	assert_array_equal(model1.predict([[1, None]]), [[1, 0]])
	assert_array_equal(model1.predict([[None, 1]]), [[1, 1]])

	assert_array_equal(model2.predict([[1, 0, None, None]]), [[1, 0, 0, 1]])
	assert_array_equal(model2.predict([[0, 0, None, None]]), [[0, 0, 0, 1]])
	assert_array_equal(model2.predict([[None, 1, None, None]]), [[1, 1, 1, 1]])
	assert_array_equal(model2.predict([[None, 1, 1, None]]), [[1, 1, 1, 1]])
	assert_array_equal(model2.predict([[None, 1, None, 0]]), [[1, 1, 0, 0]])

	assert_array_equal(model3.predict([[1, 0, None, 1, None]]), 
		[[1, 0, 0, 1, 1]])
	assert_array_equal(model3.predict([[None, 0, None, 1, None]]), 
		[[1, 0, 0, 1, 1]])
	assert_array_equal(model3.predict([[1, 0, None, 0, 1]]), 
		[[1, 0, 0, 0, 1]])
	assert_array_equal(model3.predict([[None, None, None, None, None]]), 
		[[1, 1, 1, 1, 1]])
	assert_array_equal(model3.predict([[None, None, None, None, 1]]), 
		[[1, 1, 0, 1, 1]])

@with_setup(setup_markov_network_str, teardown)
def test_predict_str():
	assert_array_equal(model1.predict([['1', None]]), [['1', '0']])
	assert_array_equal(model1.predict([[None, '1']]), [['1', '1']])

	assert_array_equal(model2.predict([['1', '0', None, None]]), [['1', '0', '0', '1']])
	assert_array_equal(model2.predict([['0', '0', None, None]]), [['0', '0', '0', '1']])
	assert_array_equal(model2.predict([[None, '1', None, None]]), [['1', '1', '1', '1']])
	assert_array_equal(model2.predict([[None, '1', '1', None]]), [['1', '1', '1', '1']])
	assert_array_equal(model2.predict([[None, '1', None, '0']]), [['1', '1', '0', '0']])

	assert_array_equal(model3.predict([['1', '0', None, '1', None]]), 
		[['1', '0', '0', '1', '1']])
	assert_array_equal(model3.predict([[None, '0', None, '1', None]]), 
		[['1', '0', '0', '1', '1']])
	assert_array_equal(model3.predict([['1', '0', None, '0', '1']]), 
		[['1', '0', '0', '0', '1']])
	assert_array_equal(model3.predict([[None, None, None, None, None]]), 
		[['1', '1', '1', '1', '1']])
	assert_array_equal(model3.predict([[None, None, None, None, '1']]), 
		[['1', '1', '0', '1', '1']])

@with_setup(setup_markov_network_bool, teardown)
def test_predict_bool():
	assert_array_equal(model1.predict([[True, None]]), [[True, False]])
	assert_array_equal(model1.predict([[None, True]]), [[True, True]])

	assert_array_equal(model2.predict([[True, False, None, None]]), [[True, False, False, True]])
	assert_array_equal(model2.predict([[False, False, None, None]]), [[False, False, False, True]])
	assert_array_equal(model2.predict([[None, True, None, None]]), [[True, True, True, True]])
	assert_array_equal(model2.predict([[None, True, True, None]]), [[True, True, True, True]])
	assert_array_equal(model2.predict([[None, True, None, False]]), [[True, True, False, False]])

	assert_array_equal(model3.predict([[True, False, None, True, None]]), 
		[[True, False, False, True, True]])
	assert_array_equal(model3.predict([[None, False, None, True, None]]), 
		[[True, False, False, True, True]])
	assert_array_equal(model3.predict([[True, False, None, False, True]]), 
		[[True, False, False, False, True]])
	assert_array_equal(model3.predict([[None, None, None, None, None]]), 
		[[True, True, True, True, True]])
	assert_array_equal(model3.predict([[None, None, None, None, True]]), 
		[[True, True, False, True, True]])


@with_setup(setup_markov_network_mixed, teardown)
def test_predict_mixed():
	assert_array_equal(model1.predict([[True, None]]), 
		numpy.array([[True, 'blue']], dtype=object))
	assert_array_equal(model1.predict([[None, 'red']]), 
		numpy.array([[True, 'red']], dtype=object))

	assert_array_equal(model2.predict([[True, 'blue', None, None]]), 
		numpy.array([[True, 'blue', False, 1]], dtype=object))
	assert_array_equal(model2.predict([[False, 'blue', None, None]]), 
		numpy.array([[False, 'blue', False, 1]], dtype=object))
	assert_array_equal(model2.predict([[None, 'red', None, None]]),
		numpy.array([[True, 'red', True, 1]], dtype=object))
	assert_array_equal(model2.predict([[None, 'red', True, None]]),
		numpy.array([[True, 'red', True, 1]], dtype=object))
	assert_array_equal(model2.predict([[None, 'red', None, 0]]),
		numpy.array([[True, 'red', False, 0]], dtype=object))

	assert_array_equal(model3.predict([[True, 'blue', None, 1, None]]), 
		numpy.array([[True, 'blue', False, 1, 'b']], dtype=object))



from pomegranate import *

from nose.tools import assert_equal

def setup():
    '''
    No setup or teardown needs to be done in this case.
    '''
    pass


def teardown():
    '''
    No setup or teardown needs to be done in this case.
    '''
    pass

def test_json():
    d1 = DiscreteDistribution({"A": 0.1, "B": 0.9})
    d2 = ConditionalProbabilityTable(
        [["A", "A", 0.1], ["A", "B", 0.9], ["B", "A", 0.7], ["B", "B", 0.3]],
        [d1])
    bayes_net = BayesianNetwork("test")
    s1 = Node(d1, name="d1")
    s2 = Node(d2, name="d2")
    bayes_net.add_states(s1, s2)
    bayes_net.add_edge(s1, s2)
    bayes_net.bake()
    fg = bayes_net.graph
    also_fg = FactorGraph.from_json(fg.to_json())
    assert_equal(fg.to_json(), also_fg.to_json())
    assert_equal(len(fg.edges), len(also_fg.edges))
    for e1, e2 in zip(fg.edges, also_fg.edges):
        assert_equal(e1[0], e2[0])
        assert_equal(e1[1], e2[1])

def test_robust_json():
    d1 = DiscreteDistribution({"A": 0.1, "B": 0.9})
    d2 = ConditionalProbabilityTable(
        [["A", "A", 0.1], ["A", "B", 0.9], ["B", "A", 0.7], ["B", "B", 0.3]],
        [d1])
    bayes_net = BayesianNetwork("test")
    s1 = Node(d1, name="d1")
    s2 = Node(d2, name="d2")
    bayes_net.add_states(s1, s2)
    bayes_net.add_edge(s1, s2)
    bayes_net.bake()
    fg = bayes_net.graph
    also_fg = from_json(fg.to_json())
    assert_equal(fg.to_json(), also_fg.to_json())
    assert_equal(len(fg.edges), len(also_fg.edges))
    for e1, e2 in zip(fg.edges, also_fg.edges):
        assert_equal(e1[0], e2[0])
        assert_equal(e1[1], e2[1])



from __future__ import (division)

from pomegranate import *
from pomegranate.io import DataGenerator
from pomegranate.io import DataFrameGenerator

from nose.tools import with_setup
from nose.tools import assert_almost_equal
from nose.tools import assert_equal
from nose.tools import assert_not_equal
from nose.tools import assert_less_equal
from nose.tools import assert_raises
from nose.tools import assert_true
from numpy.testing import assert_array_almost_equal

import pandas
import random
import pickle
import numpy as np

nan = numpy.nan

def setup_multivariate_gaussian():
	mu, cov = [0, 0, 0], numpy.eye(3)
	d1 = MultivariateGaussianDistribution(mu, cov)

	mu, cov = [2, 2, 2], numpy.eye(3)
	d2 = MultivariateGaussianDistribution(mu, cov)

	global model
	model = BayesClassifier([d1, d2])

	global X
	X = numpy.array([[ 0.3,  0.5,  0.1],
					 [ 0.8,  1.4,  0.5],
					 [ 1.4,  2.6,  1.8],
					 [ 4.2,  3.3,  3.7],
					 [ 2.6,  3.6,  3.3],
					 [ 3.1,  2.2,  1.7],
					 [ 1.8,  2.2,  1.8],
					 [-1.2, -1.8, -1.5],
					 [-1.8,  0.3,  0.5],
					 [ 0.7, -1.3, -0.1]])

	global y
	y = [0, 0, 0, 1, 1, 1, 1, 0, 0, 0]

	global X_nan
	X_nan = numpy.array([[ 0.3,  nan,  0.1],
		     			 [ nan,  1.4,  nan],
			     		 [ 1.4,  2.6,  nan],
				    	 [ nan,  nan,  nan],
					     [ nan,  3.6,  3.3],
					     [ 3.1,  nan,  1.7],
						 [ nan,  nan,  1.8],
						 [-1.2, -1.8, -1.5],
						 [ nan,  0.3,  0.5],
						 [ nan, -1.3,  nan]])


def setup_multivariate_mixed():
	mu, cov = [0, 0, 0], numpy.eye(3)
	d1 = MultivariateGaussianDistribution(mu, cov)

	d21 = ExponentialDistribution(5)
	d22 = LogNormalDistribution(0.2, 0.8)
	d23 = PoissonDistribution(3)
	d2 = IndependentComponentsDistribution([d21, d22, d23])

	global model
	model = BayesClassifier([d1, d2])

	global X
	X = numpy.array([[ 0.3,  0.5,  0.1],
					 [ 0.8,  1.4,  0.5],
					 [ 1.4,  2.6,  1.8],
					 [ 4.2,  3.3,  3.7],
					 [ 2.6,  3.6,  3.3],
					 [ 3.1,  2.2,  1.7],
					 [ 1.8,  2.2,  1.8],
					 [ 1.2,  1.8,  1.5],
					 [ 1.8,  0.3,  0.5],
					 [ 0.7,  1.3,  0.1]])

	global y
	y = [0, 0, 0, 1, 1, 1, 1, 0, 0, 0]

	global X_nan
	X_nan = numpy.array([[ 0.3,  nan,  0.1],
		     			 [ nan,  1.4,  nan],
			     		 [ 1.4,  2.6,  nan],
				    	 [ nan,  nan,  nan],
					     [ nan,  3.6,  3.3],
					     [ 3.1,  nan,  1.7],
						 [ nan,  nan,  1.8],
						 [ 1.2,  1.8,  1.5],
						 [ nan,  0.3,  0.5],
						 [ nan,  1.3,  nan]])


def setup_hmm():
	global model
	global hmm1
	global hmm2
	global hmm3

	rigged = State( DiscreteDistribution({ 'H': 0.8, 'T': 0.2 }) )
	unrigged = State( DiscreteDistribution({ 'H': 0.5, 'T':0.5 }) )

	hmm1 = HiddenMarkovModel()
	hmm1.start = rigged
	hmm1.add_transition(rigged, rigged, 1)
	hmm1.bake()

	hmm2 = HiddenMarkovModel()
	hmm2.start = unrigged
	hmm2.add_transition(unrigged, unrigged, 1)
	hmm2.bake()

	hmm3 = HiddenMarkovModel()
	hmm3.add_transition(hmm3.start, unrigged, 0.5)
	hmm3.add_transition(hmm3.start, rigged, 0.5)
	hmm3.add_transition(rigged, rigged, 0.5)
	hmm3.add_transition(rigged, unrigged, 0.5)
	hmm3.add_transition(unrigged, rigged, 0.5)
	hmm3.add_transition(unrigged, unrigged, 0.5)
	hmm3.bake()

	model = BayesClassifier([hmm1, hmm2, hmm3])


def setup_multivariate():
	pass


def teardown():
	pass


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_initialization():
	assert_equal(model.d, 3)
	assert_equal(model.n, 2)
	assert_equal(model.is_vl_, False)

@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_initialization():
	assert_equal(model.d, 3)
	assert_equal(model.n, 2)
	assert_equal(model.is_vl_, False)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_predict_log_proba():
	y_hat = model.predict_log_proba(X)
	y = [[ -1.48842547e-02,  -4.21488425e+00],
		 [ -4.37487950e-01,  -1.03748795e+00],
		 [ -5.60369104e+00,  -3.69104343e-03],
		 [ -1.64000001e+01,  -7.54345812e-08],
		 [ -1.30000023e+01,  -2.26032685e-06],
		 [ -8.00033541e+00,  -3.35406373e-04],
		 [ -5.60369104e+00,  -3.69104343e-03],
		 [ -3.05902274e-07,  -1.50000003e+01],
		 [ -3.35406373e-04,  -8.00033541e+00],
		 [ -6.11066022e-04,  -7.40061107e+00]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_predict_log_proba():
	y_hat = model.predict_log_proba(X)
	y = [[ -5.03107596e-01,  -9.27980626e-01],
		 [ -1.86355320e-01,  -1.77183117e+00],
		 [ -5.58542088e-01,  -8.48731256e-01],
		 [ -7.67315597e-01,  -6.24101927e-01],
		 [ -2.32860808e+00,  -1.02510436e-01],
		 [ -3.06641866e-03,  -5.78877778e+00],
		 [ -9.85292840e-02,  -2.36626165e+00],
		 [ -2.61764180e-01,  -1.46833995e+00],
		 [ -2.01640009e-03,  -6.20744952e+00],
		 [ -1.47371167e-01,  -1.98758175e+00]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_nan_predict_log_proba():
	y_hat = model.predict_log_proba(X_nan)
	y = [[ -3.99533332e-02,  -3.23995333e+00],
		 [ -1.17110067e+00,  -3.71100666e-01],
		 [ -4.01814993e+00,  -1.81499279e-02],
		 [ -6.93147181e-01,  -6.93147181e-01],
		 [ -9.80005545e+00,  -5.54500620e-05],
		 [ -5.60369104e+00,  -3.69104343e-03],
		 [ -1.78390074e+00,  -1.83900741e-01],
		 [ -3.05902274e-07,  -1.50000003e+01],
		 [ -8.68361522e-02,  -2.48683615e+00],
		 [ -1.00016521e-02,  -4.61000165e+00]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_nan_predict_log_proba():
	y_hat = model.predict_log_proba(X_nan)
	y = [[ -3.57980882e-01,  -1.20093223e+00],
		 [ -1.20735130e+00,  -3.55230506e-01],
		 [ -2.43174286e-01,  -1.53310132e+00],
		 [ -6.93147181e-01,  -6.93147181e-01],
		 [ -9.31781101e+00,  -8.98143220e-05],
		 [ -6.29755079e-04,  -7.37049444e+00],
		 [ -1.31307006e+00,  -3.13332194e-01],
		 [ -2.61764180e-01,  -1.46833995e+00],
		 [ -2.29725479e-01,  -1.58353505e+00],
		 [ -1.17299253e+00,  -3.70251760e-01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_predict_log_proba_parallel():
	y_hat = model.predict_log_proba(X, n_jobs=2)
	y = [[ -1.48842547e-02,  -4.21488425e+00],
		 [ -4.37487950e-01,  -1.03748795e+00],
		 [ -5.60369104e+00,  -3.69104343e-03],
		 [ -1.64000001e+01,  -7.54345812e-08],
		 [ -1.30000023e+01,  -2.26032685e-06],
		 [ -8.00033541e+00,  -3.35406373e-04],
		 [ -5.60369104e+00,  -3.69104343e-03],
		 [ -3.05902274e-07,  -1.50000003e+01],
		 [ -3.35406373e-04,  -8.00033541e+00],
		 [ -6.11066022e-04,  -7.40061107e+00]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_predict_log_proba_parallel():
	y_hat = model.predict_log_proba(X, n_jobs=2)
	y = [[ -5.03107596e-01,  -9.27980626e-01],
		 [ -1.86355320e-01,  -1.77183117e+00],
		 [ -5.58542088e-01,  -8.48731256e-01],
		 [ -7.67315597e-01,  -6.24101927e-01],
		 [ -2.32860808e+00,  -1.02510436e-01],
		 [ -3.06641866e-03,  -5.78877778e+00],
		 [ -9.85292840e-02,  -2.36626165e+00],
		 [ -2.61764180e-01,  -1.46833995e+00],
		 [ -2.01640009e-03,  -6.20744952e+00],
		 [ -1.47371167e-01,  -1.98758175e+00]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_predict_proba():
	y_hat = model.predict_proba(X)
	y =	[[  9.85225968e-01,   1.47740317e-02],
		 [  6.45656306e-01,   3.54343694e-01],
		 [  3.68423990e-03,   9.96315760e-01],
		 [  7.54345778e-08,   9.99999925e-01],
		 [  2.26032430e-06,   9.99997740e-01],
		 [  3.35350130e-04,   9.99664650e-01],
		 [  3.68423990e-03,   9.96315760e-01],
		 [  9.99999694e-01,   3.05902227e-07],
		 [  9.99664650e-01,   3.35350130e-04],
		 [  9.99389121e-01,   6.10879359e-04]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_predict_proba():
	y_hat = model.predict_proba(X)
	y = [[ 0.60464873,  0.39535127],
		 [ 0.82997863,  0.17002137],
		 [ 0.57204244,  0.42795756],
		 [ 0.46425765,  0.53574235],
		 [ 0.09743127,  0.90256873],
		 [ 0.99693828,  0.00306172],
		 [ 0.90616916,  0.09383084],
		 [ 0.76969251,  0.23030749],
		 [ 0.99798563,  0.00201437],
		 [ 0.86297361,  0.13702639]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_nan_predict_proba():
	y_hat = model.predict_proba(X_nan)
	y = [[  9.60834277e-01,   3.91657228e-02],
		 [  3.10025519e-01,   6.89974481e-01],
		 [  1.79862100e-02,   9.82013790e-01],
		 [  5.00000000e-01,   5.00000000e-01],
		 [  5.54485247e-05,   9.99944551e-01],
		 [  3.68423990e-03,   9.96315760e-01],
		 [  1.67981615e-01,   8.32018385e-01],
		 [  9.99999694e-01,   3.05902227e-07],
		 [  9.16827304e-01,   8.31726965e-02],
		 [  9.90048198e-01,   9.95180187e-03]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_nan_predict_proba():
	y_hat = model.predict_proba(X_nan)
	y = [[  6.99086440e-01,   3.00913560e-01],
		 [  2.98988163e-01,   7.01011837e-01],
		 [  7.84134838e-01,   2.15865162e-01],
		 [  5.00000000e-01,   5.00000000e-01],
		 [  8.98102888e-05,   9.99910190e-01],
		 [  9.99370443e-01,   6.29556825e-04],
		 [  2.68992964e-01,   7.31007036e-01],
		 [  7.69692511e-01,   2.30307489e-01],
		 [  7.94751748e-01,   2.05248252e-01],
		 [  3.09439547e-01,   6.90560453e-01]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_predict_proba_parallel():
	y_hat = model.predict_proba(X, n_jobs=2)
	y = [[  9.85225968e-01,   1.47740317e-02],
		 [  6.45656306e-01,   3.54343694e-01],
		 [  3.68423990e-03,   9.96315760e-01],
		 [  7.54345778e-08,   9.99999925e-01],
		 [  2.26032430e-06,   9.99997740e-01],
		 [  3.35350130e-04,   9.99664650e-01],
		 [  3.68423990e-03,   9.96315760e-01],
		 [  9.99999694e-01,   3.05902227e-07],
		 [  9.99664650e-01,   3.35350130e-04],
		 [  9.99389121e-01,   6.10879359e-04]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_predict_proba_parallel():
	y_hat = model.predict_proba(X, n_jobs=2)
	y = [[ 0.60464873,  0.39535127],
		 [ 0.82997863,  0.17002137],
		 [ 0.57204244,  0.42795756],
		 [ 0.46425765,  0.53574235],
		 [ 0.09743127,  0.90256873],
		 [ 0.99693828,  0.00306172],
		 [ 0.90616916,  0.09383084],
		 [ 0.76969251,  0.23030749],
		 [ 0.99798563,  0.00201437],
		 [ 0.86297361,  0.13702639]]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_predict():
	y_hat = model.predict(X)
	y = [0, 0, 1, 1, 1, 1, 1, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_predict():
	y_hat = model.predict(X)
	y = [0, 0, 0, 1, 1, 0, 0, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_nan_predict():
	y_hat = model.predict(X_nan)
	y = [0, 1, 1, 0, 1, 1, 1, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_nan_predict():
	y_hat = model.predict(X_nan)
	y = [0, 1, 0, 0, 1, 0, 1, 0, 0, 1]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_predict_parallel():
	y_hat = model.predict(X, n_jobs=2)
	y = [0, 0, 1, 1, 1, 1, 1, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_predict_parallel():
	y_hat = model.predict(X, n_jobs=2)
	y = [0, 0, 0, 1, 1, 0, 0, 0, 0, 0]

	assert_array_almost_equal(y, y_hat)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_fit_parallel():
	model.fit(X, y, n_jobs=2)

	mu1 = model.distributions[0].parameters[0]
	cov1 = model.distributions[0].parameters[1]
	mu1_t = [0.03333333, 0.28333333, 0.21666666]
	cov1_t = [[1.3088888, 0.9272222, 0.6227777],
			  [0.9272222, 2.2513888, 1.3402777],
			  [0.6227777, 1.3402777, 0.9547222]]

	mu2 = model.distributions[1].parameters[0]
	cov2 = model.distributions[1].parameters[1]
	mu2_t = [2.925, 2.825, 2.625]
	cov2_t = [[0.75687499, 0.23687499, 0.4793750],
			  [0.23687499, 0.40187499, 0.5318749],
			  [0.47937500, 0.53187499, 0.7868750]]


	assert_array_almost_equal(mu1, mu1_t)
	assert_array_almost_equal(cov1, cov1_t)
	assert_array_almost_equal(mu2, mu2_t)
	assert_array_almost_equal(cov2, cov2_t)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_fit_parallel():
	model.fit(X, y, n_jobs=2)

	mu1 = model.distributions[0].parameters[0]
	cov1 = model.distributions[0].parameters[1]
	mu1_t = [1.033333, 1.3166667, 0.75]
	cov1_t = [[0.242222, 0.0594444, 0.178333],
			  [0.059444, 0.5980555, 0.414166],
			  [0.178333, 0.4141666, 0.439166]]

	d21 = model.distributions[1].distributions[0]
	d22 = model.distributions[1].distributions[1]
	d23 = model.distributions[1].distributions[2]

	assert_array_almost_equal(mu1, mu1_t)
	assert_array_almost_equal(cov1, cov1_t)
	assert_array_almost_equal(d21.parameters, [0.34188034])
	assert_array_almost_equal(d22.parameters, [1.01294275, 0.22658346])
	assert_array_almost_equal(d23.parameters, [2.625])


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_from_samples():
	model = BayesClassifier.from_samples(MultivariateGaussianDistribution, X, y)

	mu1 = model.distributions[0].parameters[0]
	cov1 = model.distributions[0].parameters[1]
	mu1_t = [0.03333333, 0.2833333, 0.21666666]
	cov1_t = [[1.308888888, 0.9272222222, 0.6227777777],
			  [0.927222222, 2.251388888, 1.340277777],
			  [0.622777777, 1.340277777, 0.9547222222]]

	mu2 = model.distributions[1].parameters[0]
	cov2 = model.distributions[1].parameters[1]
	mu2_t = [2.925, 2.825, 2.625]
	cov2_t = [[0.75687500, 0.23687499, 0.47937500],
			  [0.23687499, 0.40187499, 0.53187499],
			  [0.47937500, 0.53187499, 0.78687500]]

	assert_array_almost_equal(mu1, mu1_t)
	assert_array_almost_equal(cov1, cov1_t)
	assert_array_almost_equal(mu2, mu2_t)
	assert_array_almost_equal(cov2, cov2_t)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_pickle():
	model2 = pickle.loads(pickle.dumps(model))

	assert_true(isinstance(model2, BayesClassifier))
	assert_true(isinstance(model2.distributions[0], MultivariateGaussianDistribution))
	assert_true(isinstance(model2.distributions[1], MultivariateGaussianDistribution))
	assert_array_almost_equal(model.weights, model2.weights)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_pickle():
	model2 = pickle.loads(pickle.dumps(model))

	assert_true(isinstance(model2, BayesClassifier))
	assert_true(isinstance(model2.distributions[0], MultivariateGaussianDistribution))
	assert_true(isinstance(model2.distributions[1], IndependentComponentsDistribution))
	assert_array_almost_equal(model.weights, model2.weights)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_to_json():
	model2 = BayesClassifier.from_json(model.to_json())

	assert_true(isinstance(model2, BayesClassifier))
	assert_true(isinstance(model2.distributions[0], MultivariateGaussianDistribution))
	assert_true(isinstance(model2.distributions[1], MultivariateGaussianDistribution))
	assert_array_almost_equal(model.weights, model2.weights)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_to_json():
	model2 = BayesClassifier.from_json(model.to_json())

	assert_true(isinstance(model2, BayesClassifier))
	assert_true(isinstance(model2.distributions[0], MultivariateGaussianDistribution))
	assert_true(isinstance(model2.distributions[1], IndependentComponentsDistribution))
	assert_array_almost_equal(model.weights, model2.weights)


@with_setup(setup_multivariate_gaussian, teardown)
def test_bc_multivariate_gaussian_robust_from_json():
	model2 = from_json(model.to_json())

	assert_true(isinstance(model2, BayesClassifier))
	assert_true(isinstance(model2.distributions[0], MultivariateGaussianDistribution))
	assert_true(isinstance(model2.distributions[1], MultivariateGaussianDistribution))
	assert_array_almost_equal(model.weights, model2.weights)


@with_setup(setup_multivariate_mixed, teardown)
def test_bc_multivariate_mixed_robust_from_json():
	model2 = from_json(model.to_json())

	assert_true(isinstance(model2, BayesClassifier))
	assert_true(isinstance(model2.distributions[0], MultivariateGaussianDistribution))
	assert_true(isinstance(model2.distributions[1], IndependentComponentsDistribution))
	assert_array_almost_equal(model.weights, model2.weights)


@with_setup(setup_hmm, teardown)
def test_model():
	assert_almost_equal(hmm1.log_probability(list('H')), -0.2231435513142097 )
	assert_almost_equal(hmm1.log_probability(list('T')), -1.6094379124341003 )
	assert_almost_equal(hmm1.log_probability(list('HHHH')), -0.8925742052568388 )
	assert_almost_equal(hmm1.log_probability(list('THHH')), -2.2788685663767296 )
	assert_almost_equal(hmm1.log_probability(list('TTTT')), -6.437751649736401 )

	assert_almost_equal(hmm2.log_probability(list('H')), -0.6931471805599453 )
	assert_almost_equal(hmm2.log_probability(list('T')), -0.6931471805599453 )
	assert_almost_equal(hmm2.log_probability(list('HHHH')), -2.772588722239781 )
	assert_almost_equal(hmm2.log_probability(list('THHH')), -2.772588722239781 )
	assert_almost_equal(hmm2.log_probability(list('TTTT')), -2.772588722239781 )

	assert_almost_equal(hmm3.log_probability(list('H')), -0.43078291609245417)
	assert_almost_equal(hmm3.log_probability(list('T')), -1.0498221244986776)
	assert_almost_equal(hmm3.log_probability(list('HHHH')), -1.7231316643698167)
	assert_almost_equal(hmm3.log_probability(list('THHH')), -2.3421708727760397)
	assert_almost_equal(hmm3.log_probability(list('TTTT')), -4.1992884979947105)
	assert_almost_equal(hmm3.log_probability(list('THTHTHTHTHTH')), -8.883630243546788)
	assert_almost_equal(hmm3.log_probability(list('THTHHHHHTHTH')), -7.645551826734343)

	assert_equal(model.d, 1)


@with_setup(setup_hmm, teardown)
def test_hmm_log_proba():
	logs = model.predict_log_proba(np.array([list('H'), list('THHH'), list('TTTT'), list('THTHTHTHTHTH'), list('THTHHHHHTHTH')]))

	assert_almost_equal(logs[0][0], -0.89097292388986515)
	assert_almost_equal(logs[0][1], -1.3609765531356006)
	assert_almost_equal(logs[0][2], -1.0986122886681096)

	assert_almost_equal(logs[1][0], -0.93570553121744293)
	assert_almost_equal(logs[1][1], -1.429425687080494)
	assert_almost_equal(logs[1][2], -0.9990078376167526)

	assert_almost_equal(logs[2][0], -3.9007882563128864)
	assert_almost_equal(logs[2][1], -0.23562532881626597)
	assert_almost_equal(logs[2][2], -1.6623251045711958)

	assert_almost_equal(logs[3][0], -3.1703366478831185)
	assert_almost_equal(logs[3][1], -0.49261403211260379)
	assert_almost_equal(logs[3][2], -1.058478108940049)

	assert_almost_equal(logs[4][0], -1.3058441172130273)
	assert_almost_equal(logs[4][1], -1.4007102236822906)
	assert_almost_equal(logs[4][2], -0.7284958836972919)


@with_setup(setup_hmm, teardown)
def test_hmm_proba():
	probs = model.predict_proba(np.array([list('H'), list('THHH'), list('TTTT'), list('THTHTHTHTHTH'), list('THTHHHHHTHTH')]))

	assert_almost_equal(probs[0][0], 0.41025641025641024)
	assert_almost_equal(probs[0][1], 0.25641025641025639)
	assert_almost_equal(probs[0][2], 0.33333333333333331)

	assert_almost_equal(probs[1][0], 0.39230898163446098)
	assert_almost_equal(probs[1][1], 0.23944639992337707)
	assert_almost_equal(probs[1][2], 0.36824461844216183)

	assert_almost_equal(probs[2][0], 0.020225961918306088)
	assert_almost_equal(probs[2][1], 0.79007663743383105)
	assert_almost_equal(probs[2][2], 0.18969740064786292)

	assert_almost_equal(probs[3][0], 0.041989459861032523)
	assert_almost_equal(probs[3][1], 0.61102706038265642)
	assert_almost_equal(probs[3][2], 0.346983479756311)

	assert_almost_equal(probs[4][0], 0.27094373022369794)
	assert_almost_equal(probs[4][1], 0.24642188711704707)
	assert_almost_equal(probs[4][2], 0.48263438265925512)


@with_setup(setup_hmm, teardown)
def test_hmm_prediction():
	predicts = model.predict(np.array([list('H'), list('THHH'), list('TTTT'), list('THTHTHTHTHTH'), list('THTHHHHHTHTH')]))

	assert_equal(predicts[0], 0)
	assert_equal(predicts[1], 0)
	assert_equal(predicts[2], 1)
	assert_equal(predicts[3], 1)
	assert_equal(predicts[4], 2)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_log_probability():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))

	logp1 = model.log_probability(X)
	logp2 = model.log_probability(X2)
	logp3 = model.log_probability(X3)

	assert_array_almost_equal(logp1, logp2)
	assert_array_almost_equal(logp1, logp3)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_predict():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))

	y_hat1 = model.predict(X)
	y_hat2 = model.predict(X2)
	y_hat3 = model.predict(X3)

	assert_array_almost_equal(y_hat1, y_hat2)
	assert_array_almost_equal(y_hat1, y_hat3)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_predict_proba():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))

	y_hat1 = model.predict_proba(X)
	y_hat2 = model.predict_proba(X2)
	y_hat3 = model.predict_proba(X3)

	assert_array_almost_equal(y_hat1, y_hat2)
	assert_array_almost_equal(y_hat1, y_hat3)

@with_setup(setup_multivariate_gaussian, teardown)
def test_io_predict_log_proba():
	X2 = DataGenerator(X)
	X3 = DataFrameGenerator(pandas.DataFrame(X))

	y_hat1 = model.predict_log_proba(X)
	y_hat2 = model.predict_log_proba(X2)
	y_hat3 = model.predict_log_proba(X3)

	assert_array_almost_equal(y_hat1, y_hat2)
	assert_array_almost_equal(y_hat1, y_hat3)

def test_io_fit():
	X = numpy.random.randn(100, 5) + 0.5
	weights = numpy.abs(numpy.random.randn(100))
	y = numpy.random.randint(2, size=100)
	data_generator = DataGenerator(X, weights, y)

	mu1 = numpy.array([0, 0, 0, 0, 0])
	mu2 = numpy.array([1, 1, 1, 1, 1])
	cov = numpy.eye(5)

	d1 = MultivariateGaussianDistribution(mu1, cov)
	d2 = MultivariateGaussianDistribution(mu2, cov)
	bc1 = BayesClassifier([d1, d2])
	bc1.fit(X, y, weights)

	d1 = MultivariateGaussianDistribution(mu1, cov)
	d2 = MultivariateGaussianDistribution(mu2, cov)
	bc2 = BayesClassifier([d1, d2])
	bc2.fit(data_generator)

	logp1 = bc1.log_probability(X)
	logp2 = bc2.log_probability(X)

	assert_array_almost_equal(logp1, logp2)

def test_io_from_samples():
	X = numpy.random.randn(100, 5) + 0.5
	weights = numpy.abs(numpy.random.randn(100))
	y = numpy.random.randint(2, size=100)
	data_generator = DataGenerator(X, weights, y)

	d = MultivariateGaussianDistribution

	bc1 = BayesClassifier.from_samples(d, X=X, y=y, weights=weights)
	bc2 = BayesClassifier.from_samples(d, X=data_generator)

	logp1 = bc1.log_probability(X)
	logp2 = bc2.log_probability(X)

	assert_array_almost_equal(logp1, logp2)


from pomegranate import *
from pomegranate.io import DataGenerator
from pomegranate.io import SequenceGenerator
from pomegranate.io import DataFrameGenerator

from nose.tools import with_setup
from nose.tools import assert_true
from nose.tools import assert_equal
from nose.tools import assert_raises
from numpy.testing import assert_almost_equal
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

import random
import numpy
import pandas

numpy.random.seed(0)
random.seed(0)

nan = numpy.nan

def test_io_datagenerator_shape():
	X = numpy.random.randn(500, 13)
	data = DataGenerator(X)

	assert_array_equal(data.shape, X.shape)

def test_io_datagenerator_classes_fail():
	X = numpy.random.randn(500, 13)
	data = DataGenerator(X)

	assert_raises(ValueError, lambda data: data.classes, data)

def test_io_datagenerator_classes():
	X = numpy.random.randn(500, 13)
	y = numpy.random.randint(5, size=500)
	data = DataGenerator(X, y=y)

	assert_array_equal(data.classes, [0, 1, 2, 3, 4])

def test_io_datagenerator_x_batches():
	X = numpy.random.randn(500, 13)
	w = numpy.ones(500)

	data = DataGenerator(X)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X, X_)
	assert_almost_equal(w, w_)

	data = DataGenerator(X, batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X, X_)
	assert_almost_equal(w, w_)

	data = DataGenerator(X, batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X, X_)
	assert_almost_equal(w, w_)

	data = DataGenerator(X, batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X, X_)
	assert_almost_equal(w, w_)

def test_io_datagenerator_w_batches():
	X = numpy.random.randn(500, 13)
	w = numpy.abs(numpy.random.randn(500))

	data = DataGenerator(X, w)
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(w, w_)

	data = DataGenerator(X, w, batch_size=123)
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(w, w_)

	data = DataGenerator(X, w, batch_size=1)
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(w, w_)

	data = DataGenerator(X, w, batch_size=506)
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(w, w_)

def test_io_datagenerator_y_batches():
	X = numpy.random.randn(500, 13)
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500)

	data = DataGenerator(X, y=y)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(y, y_)

	data = DataGenerator(X, y=y, batch_size=123)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(y, y_)

	data = DataGenerator(X, y=y, batch_size=1)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(y, y_)

	data = DataGenerator(X, y=y, batch_size=506)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(y, y_)

def test_io_datagenerator_wy_batches():
	X = numpy.random.randn(500, 13)
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500)

	data = DataGenerator(X, w, y)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataGenerator(X, w, y, batch_size=123)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataGenerator(X, w, y, batch_size=1)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataGenerator(X, w, y, batch_size=506)
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

def test_io_datagenerator_wy_unlabeled():
	X = numpy.random.randn(500, 13)
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500) - 1

	data = DataGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.unlabeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.unlabeled_batches()])

	assert_true(X.shape[0] > X_.shape[0])
	assert_almost_equal(X[y == -1], X_)
	assert_almost_equal(w[y == -1], w_)

def test_io_datagenerator_wy_labeled():
	X = numpy.random.randn(500, 13)
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500) - 1

	data = DataGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.labeled_batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.labeled_batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.labeled_batches()])

	assert_true(X.shape[0] > X_.shape[0])
	assert_almost_equal(X[y != -1], X_)
	assert_almost_equal(y[y != -1], y_)
	assert_almost_equal(w[y != -1], w_)

def test_io_seqgenerator_uni_shape():
	X = [numpy.random.randn(15) for i in range(500)] 
	data = SequenceGenerator(X)

	assert_array_equal(data.shape, (len(X), 1))

def test_io_seqgenerator_symbol_shape():
	X = [[numpy.random.choice(['A', 'C', 'G', 'T']) for j in range(15)] for i in range(500)] 
	data = SequenceGenerator(X)

	assert_array_equal(data.shape, (len(X), 1))

def test_io_seqgenerator_shape():
	X = [numpy.random.randn(15, 8) for i in range(500)] 
	data = SequenceGenerator(X)

	assert_array_equal(data.shape, (len(X), 8))

def test_io_seqgenerator_x_batches():
	X = [numpy.random.randn(15, 8) for i in range(500)] 
	w = numpy.ones(500)

	data = SequenceGenerator(X)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X, X_)
	assert_almost_equal(w, w_)

def test_io_seqgenerator_x_symbol_batches():
	X = [[numpy.random.choice(["A", "C", "G", "T"]) for i in range(18)] for i in range(500)] 
	w = numpy.ones(500)

	data = SequenceGenerator(X)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_array_equal(X, X_)
	assert_almost_equal(w, w_)

def test_io_seqgenerator_w_batches():
	X = [numpy.random.randn(15, 8) for i in range(500)] 
	w = numpy.abs(numpy.random.randn(500))

	data = SequenceGenerator(X, w)
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(w, w_)

def test_io_seqgenerator_w_symbol_batches():
	X = [[numpy.random.choice(["A", "C", "G", "T"]) for i in range(18)] for i in range(500)] 
	w = numpy.abs(numpy.random.randn(500))

	data = SequenceGenerator(X, w)
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(w, w_)

def test_io_seqgenerator_y_batches():
	X = [numpy.random.randn(15, 8) for i in range(500)] 
	y = [numpy.random.randint(6, size=15) for i in range(500)]

	data = SequenceGenerator(X, y=y)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_array_equal(X, X_)
	assert_almost_equal(y, y_)

def test_io_seqgenerator_y_symbol_batches():
	X = [[numpy.random.choice(["A", "C", "G", "T"]) for i in range(18)] for i in range(500)] 
	y = [numpy.random.randint(6, size=15) for i in range(500)]

	data = SequenceGenerator(X, y=y)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_array_equal(X, X_)
	assert_almost_equal(y, y_)

def test_io_seqgenerator_wy_batches():
	X = [numpy.random.randn(15, 8) for i in range(500)] 
	y = [numpy.random.randint(6, size=15) for i in range(500)]
	w = numpy.abs(numpy.random.randn(500))

	data = SequenceGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	y_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X, X_)
	assert_almost_equal(w, w_)
	assert_almost_equal(y, y_)

def test_io_seqgenerator_wy_symbol_batches():
	X = [[numpy.random.choice(["A", "C", "G", "T"]) for i in range(18)] for i in range(500)] 
	y = [numpy.random.randint(6, size=15) for i in range(500)]
	w = numpy.abs(numpy.random.randn(500))

	data = SequenceGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	y_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_array_equal(X, X_)
	assert_almost_equal(w, w_)
	assert_almost_equal(y, y_)

def test_io_seqgenerator_wy_unlabeled():
	X = [numpy.random.randn(15, 8) for i in range(500)] 
	y = [numpy.random.randint(6, size=15) for i in range(500)]
	w = numpy.abs(numpy.random.randn(500))

	idx = numpy.random.choice(500, size=100, replace=False)
	for i in idx:
		y[i] = None

	data = SequenceGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.unlabeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.unlabeled_batches()])

	assert_true(len(X) > len(X_))
	assert_true(len(w) > len(w_))

	i = 0
	for j in range(500):
		if y[j] is None:
			assert_almost_equal(X[j], X_[i])
			assert_almost_equal(w[j], w_[i])
			i += 1

def test_io_seqgenerator_wy_symbol_unlabeled():
	X = [[numpy.random.choice(["A", "C", "G", "T"]) for i in range(18)] for i in range(500)]  
	y = [numpy.random.randint(6, size=15) for i in range(500)]
	w = numpy.abs(numpy.random.randn(500))

	idx = numpy.random.choice(500, size=100, replace=False)
	for i in idx:
		y[i] = None

	data = SequenceGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.unlabeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.unlabeled_batches()])

	assert_true(len(X) > len(X_))
	assert_true(len(w) > len(w_))

	i = 0
	for j in range(500):
		if y[j] is None:
			assert_array_equal(X[j], X_[i])
			assert_almost_equal(w[j], w_[i])
			i += 1

def test_io_seqgenerator_wy_labeled():
	X = [numpy.random.randn(15, 8) for i in range(500)]
	y = [numpy.random.randint(6, size=15) for i in range(500)]
	w = numpy.abs(numpy.random.randn(500))

	idx = numpy.random.choice(500, size=100, replace=False)
	for i in idx:
		y[i] = None

	data = SequenceGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.labeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.labeled_batches()])
	y_ = numpy.concatenate([batch[2] for batch in data.labeled_batches()])

	assert_true(len(X) > len(X_))
	assert_true(len(w) > len(w_))
	assert_true(len(y) > len(y_))

	i = 0
	for j in range(500):
		if y[j] is not None:
			assert_almost_equal(X[j], X_[i])
			assert_almost_equal(w[j], w_[i])
			assert_almost_equal(y[j], y_[i])
			i += 1

def test_io_seqgenerator_wy_symbol_labeled():
	X = [[numpy.random.choice(["A", "C", "G", "T"]) for i in range(18)] for i in range(500)] 
	y = [numpy.random.randint(6, size=15) for i in range(500)]
	w = numpy.abs(numpy.random.randn(500))

	idx = numpy.random.choice(500, size=100, replace=False)
	for i in idx:
		y[i] = None

	data = SequenceGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.labeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.labeled_batches()])
	y_ = numpy.concatenate([batch[2] for batch in data.labeled_batches()])

	assert_true(len(X) > len(X_))
	assert_true(len(w) > len(w_))
	assert_true(len(y) > len(y_))

	i = 0
	for j in range(500):
		if y[j] is not None:
			assert_array_equal(X[j], X_[i])
			assert_almost_equal(w[j], w_[i])
			assert_almost_equal(y[j], y_[i])
			i += 1

def test_io_dfgenerator_shape():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	data = DataFrameGenerator(X)

	assert_array_equal(data.shape, X.shape)

def test_io_dfgenerator_classes_fail():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	data = DataFrameGenerator(X)

	assert_raises(ValueError, lambda data: data.classes, data)

def test_io_dfgenerator_numpy_classes():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	y = numpy.random.randint(5, size=500)
	data = DataFrameGenerator(X, y=y)

	assert_array_equal(data.classes, [0, 1, 2, 3, 4])

def test_io_dfgenerator_pandas_classes():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	y = pandas.Series(numpy.random.randint(5, size=500))
	data = DataFrameGenerator(X, y=y)

	assert_array_equal(data.classes, [0, 1, 2, 3, 4])

def test_io_dfgenerator_str_classes():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	X['y'] = numpy.random.randint(5, size=500)
	data = DataFrameGenerator(X, y='y')

	assert_array_equal(data.classes, [0, 1, 2, 3, 4])

def test_io_dfgenerator_x_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	w = numpy.ones(500)

	data = DataFrameGenerator(X)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

def test_io_dfgenerator_w_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	w = numpy.abs(numpy.random.randn(500))

	data = DataFrameGenerator(X, w)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, w, batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, w, batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, w, batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(w, w_)

def test_io_dfgenerator_w_str_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	X2 = X.copy()
	w = numpy.abs(numpy.random.randn(500))
	X['w'] = w

	data = DataFrameGenerator(X, 'w')
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, 'w', batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, 'w', batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, 'w', batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(w, w_)

def test_io_dfgenerator_y_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	y = numpy.random.randint(5, size=500)

	data = DataFrameGenerator(X, y=y)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)

	data = DataFrameGenerator(X, y=y, batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)

	data = DataFrameGenerator(X, y=y, batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)

	data = DataFrameGenerator(X, y=y, batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)

def test_io_dfgenerator_y_str_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	X2 = X.copy()
	y = numpy.random.randint(5, size=500)
	X['y'] = y

	data = DataFrameGenerator(X, y='y')
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)

	data = DataFrameGenerator(X, y='y', batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)

	data = DataFrameGenerator(X, y='y', batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)

	data = DataFrameGenerator(X, y='y', batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)

def test_io_dfgenerator_wy_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500)

	data = DataFrameGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, w, y, batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, w, y, batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, w, y, batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

def test_io_dfgenerator_wy_str_batches():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	X2 = X.copy()
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500)

	X['w'] = w
	X['y'] = y

	data = DataFrameGenerator(X, 'w', 'y')
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, 'w', 'y', batch_size=123)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, 'w', 'y', batch_size=1)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

	data = DataFrameGenerator(X, 'w', 'y', batch_size=506)
	X_ = numpy.concatenate([batch[0] for batch in data.batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.batches()])
	assert_almost_equal(X2.values, X_)
	assert_almost_equal(y, y_)
	assert_almost_equal(w, w_)

def test_io_dfgenerator_wy_unlabeled():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500) - 1

	data = DataFrameGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.unlabeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.unlabeled_batches()])

	assert_true(X.shape[0] > X_.shape[0])
	assert_almost_equal(X.loc[y == -1], X_)
	assert_almost_equal(w[y == -1], w_)

def test_io_dfgenerator_wy_str_unlabeled():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	X2 = X.copy()
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500) - 1

	X['w'] = w
	X['y'] = y

	data = DataFrameGenerator(X, 'w', 'y')
	X_ = numpy.concatenate([batch[0] for batch in data.unlabeled_batches()])
	w_ = numpy.concatenate([batch[1] for batch in data.unlabeled_batches()])

	assert_true(X.shape[0] > X_.shape[0])
	assert_almost_equal(X2.loc[y == -1], X_)
	assert_almost_equal(w[y == -1], w_)

def test_io_dfgenerator_wy_labeled():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500) - 1

	data = DataFrameGenerator(X, w, y)
	X_ = numpy.concatenate([batch[0] for batch in data.labeled_batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.labeled_batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.labeled_batches()])

	assert_true(X.shape[0] > X_.shape[0])
	assert_almost_equal(X.loc[y != -1], X_)
	assert_almost_equal(y[y != -1], y_)
	assert_almost_equal(w[y != -1], w_)

def test_io_dfgenerator_wy_str_labeled():
	X = pandas.DataFrame(numpy.random.randn(500, 13))
	X2 = X.copy()
	w = numpy.abs(numpy.random.randn(500))
	y = numpy.random.randint(5, size=500) - 1

	X['w'] = w
	X['y'] = y

	data = DataFrameGenerator(X, 'w', 'y')
	X_ = numpy.concatenate([batch[0] for batch in data.labeled_batches()])
	y_ = numpy.concatenate([batch[1] for batch in data.labeled_batches()])
	w_ = numpy.concatenate([batch[2] for batch in data.labeled_batches()])

	assert_true(X.shape[0] > X_.shape[0])
	assert_almost_equal(X2.loc[y != -1], X_)
	assert_almost_equal(y[y != -1], y_)
	assert_almost_equal(w[y != -1], w_)



# test_bayes_net.py
# Authors: Jacob Schreiber <jmschreiber91@gmail.com>
#          Nelson Liu <nelson@nelsonliu.me>
#
'''
These are unit tests for the Bayesian network part of pomegranate.
'''

from __future__ import division

from pomegranate import from_json
from pomegranate import DiscreteDistribution
from pomegranate import ConditionalProbabilityTable
from pomegranate import State, Node
from pomegranate import BayesianNetwork
from pomegranate.BayesianNetwork import _check_input
from pomegranate.io import DataGenerator
from pomegranate.io import DataFrameGenerator

from nose.tools import with_setup
from nose.tools import assert_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import assert_almost_equal

from networkx import DiGraph

from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal

import pandas
import random, numpy
import sys

nan = numpy.nan
numpy.random.seed(1)

datasets = [numpy.random.randint(2, size=(10, 4)),
            numpy.random.randint(2, size=(100, 5)),
            numpy.random.randint(2, size=(1000, 7)),
            numpy.random.randint(2, size=(100, 9))]

datasets_nan = []
for dataset in datasets:
    X = dataset.copy().astype('float64')
    n, d = X.shape

    idx = numpy.random.choice(n*d, replace=False, size=n*d//5)
    i, j = idx // d, idx % d
    X[i, j] = numpy.nan

    datasets_nan.append(X)

def setup_monty():
    # Build a model of the Monty Hall Problem
    global monty_network, monty_index, prize_index, guest_index

    random.seed(0)

    # Friends emissions are completely random
    guest = DiscreteDistribution({'A': 1. / 3, 'B': 1. / 3, 'C': 1. / 3})

    # The actual prize is independent of the other distributions
    prize = DiscreteDistribution({'A': 1. / 3, 'B': 1. / 3, 'C': 1. / 3})
    # Monty is dependent on both the guest and the prize.
    monty = ConditionalProbabilityTable(
        [['A', 'A', 'A', 0.0],
         ['A', 'A', 'B', 0.5],
         ['A', 'A', 'C', 0.5],
         ['A', 'B', 'A', 0.0],
         ['A', 'B', 'B', 0.0],
         ['A', 'B', 'C', 1.0],
         ['A', 'C', 'A', 0.0],
         ['A', 'C', 'B', 1.0],
         ['A', 'C', 'C', 0.0],
         ['B', 'A', 'A', 0.0],
         ['B', 'A', 'B', 0.0],
         ['B', 'A', 'C', 1.0],
         ['B', 'B', 'A', 0.5],
         ['B', 'B', 'B', 0.0],
         ['B', 'B', 'C', 0.5],
         ['B', 'C', 'A', 1.0],
         ['B', 'C', 'B', 0.0],
         ['B', 'C', 'C', 0.0],
         ['C', 'A', 'A', 0.0],
         ['C', 'A', 'B', 1.0],
         ['C', 'A', 'C', 0.0],
         ['C', 'B', 'A', 1.0],
         ['C', 'B', 'B', 0.0],
         ['C', 'B', 'C', 0.0],
         ['C', 'C', 'A', 0.5],
         ['C', 'C', 'B', 0.5],
         ['C', 'C', 'C', 0.0]], [guest, prize])

    # Make the states
    s1 = State(guest, name="guest")
    s2 = State(prize, name="prize")
    s3 = State(monty, name="monty")

    # Make the bayes net, add the states, and the conditional dependencies.
    monty_network = BayesianNetwork("test")
    monty_network.add_nodes(s1, s2, s3)
    monty_network.add_edge(s1, s3)
    monty_network.add_edge(s2, s3)
    monty_network.bake()

    monty_index = monty_network.states.index(s3)
    prize_index = monty_network.states.index(s2)
    guest_index = monty_network.states.index(s1)


def setup_titanic():
    # Build a model of the titanic disaster
    global titanic_network, passenger, gender, tclass

    # Passengers on the Titanic either survive or perish
    passenger = DiscreteDistribution({'survive': 0.6, 'perish': 0.4})

    # Gender, given survival data
    gender = ConditionalProbabilityTable(
        [['survive', 'male',   0.0],
         ['survive', 'female', 1.0],
         ['perish', 'male',    1.0],
         ['perish', 'female',  0.0]], [passenger])

    # Class of travel, given survival data
    tclass = ConditionalProbabilityTable(
        [['survive', 'first',  0.0],
         ['survive', 'second', 1.0],
         ['survive', 'third',  0.0],
         ['perish', 'first',  1.0],
         ['perish', 'second', 0.0],
         ['perish', 'third',  0.0]], [passenger])

    # State objects hold both the distribution, and a high level name.
    s1 = State(passenger, name="passenger")
    s2 = State(gender, name="gender")
    s3 = State(tclass, name="class")

    # Create the Bayesian network object with a useful name
    titanic_network = BayesianNetwork("Titanic Disaster")

    # Add the three nodes to the network
    titanic_network.add_nodes(s1, s2, s3)

    # Add transitions which represent conditional dependencies, where the
    # second node is conditionally dependent on the first node (Monty is
    # dependent on both guest and prize)
    titanic_network.add_edge(s1, s2)
    titanic_network.add_edge(s1, s3)
    titanic_network.bake()


def setup_large_monty():
    # Build the huge monty hall large_monty_network. This is an example I made
    # up with which may not exactly flow logically, but tests a varied type of
    # tables ensures heterogeneous types of data work together.
    global large_monty_network, large_monty_friend, large_monty_guest, large_monty
    global large_monty_remaining, large_monty_randomize, large_monty_prize

    # large_monty_Friend
    large_monty_friend = DiscreteDistribution({True: 0.5, False: 0.5})

    # large_monty_Guest emisisons are completely random
    large_monty_guest = ConditionalProbabilityTable(
        [[True, 'A', 0.50],
         [True, 'B', 0.25],
         [True, 'C', 0.25],
         [False, 'A', 0.0],
         [False, 'B', 0.7],
         [False, 'C', 0.3]], [large_monty_friend])

    # Number of large_monty_remaining cars
    large_monty_remaining = DiscreteDistribution({0: 0.1, 1: 0.7, 2: 0.2})

    # Whether they large_monty_randomize is dependent on the numnber of
    # large_monty_remaining cars
    large_monty_randomize = ConditionalProbabilityTable(
        [[0, True, 0.05],
         [0, False, 0.95],
         [1, True, 0.8],
         [1, False, 0.2],
         [2, True, 0.5],
         [2, False, 0.5]], [large_monty_remaining])

    # Where the large_monty_prize is depends on if they large_monty_randomize or
    # not and also the large_monty_guests large_monty_friend
    large_monty_prize = ConditionalProbabilityTable(
        [[True, True, 'A', 0.3],
         [True, True, 'B', 0.4],
         [True, True, 'C', 0.3],
         [True, False, 'A', 0.2],
         [True, False, 'B', 0.4],
         [True, False, 'C', 0.4],
         [False, True, 'A', 0.1],
         [False, True, 'B', 0.9],
         [False, True, 'C', 0.0],
         [False, False, 'A', 0.0],
         [False, False, 'B', 0.4],
         [False, False, 'C', 0.6]], [large_monty_randomize, large_monty_friend])

    # Monty is dependent on both the large_monty_guest and the large_monty_prize.
    large_monty = ConditionalProbabilityTable(
        [['A', 'A', 'A', 0.0],
         ['A', 'A', 'B', 0.5],
         ['A', 'A', 'C', 0.5],
         ['A', 'B', 'A', 0.0],
         ['A', 'B', 'B', 0.0],
         ['A', 'B', 'C', 1.0],
         ['A', 'C', 'A', 0.0],
         ['A', 'C', 'B', 1.0],
         ['A', 'C', 'C', 0.0],
         ['B', 'A', 'A', 0.0],
         ['B', 'A', 'B', 0.0],
         ['B', 'A', 'C', 1.0],
         ['B', 'B', 'A', 0.5],
         ['B', 'B', 'B', 0.0],
         ['B', 'B', 'C', 0.5],
         ['B', 'C', 'A', 1.0],
         ['B', 'C', 'B', 0.0],
         ['B', 'C', 'C', 0.0],
         ['C', 'A', 'A', 0.0],
         ['C', 'A', 'B', 1.0],
         ['C', 'A', 'C', 0.0],
         ['C', 'B', 'A', 1.0],
         ['C', 'B', 'B', 0.0],
         ['C', 'B', 'C', 0.0],
         ['C', 'C', 'A', 0.5],
         ['C', 'C', 'B', 0.5],
         ['C', 'C', 'C', 0.0]], [large_monty_guest, large_monty_prize])

    # Make the states
    s0 = State(large_monty_friend, name="large_monty_friend")
    s1 = State(large_monty_guest, name="large_monty_guest")
    s2 = State(large_monty_prize, name="large_monty_prize")
    s3 = State(large_monty, name="large_monty")
    s4 = State(large_monty_remaining, name="large_monty_remaining")
    s5 = State(large_monty_randomize, name="large_monty_randomize")

    # Make the bayes net, add the states, and the conditional dependencies.
    large_monty_network = BayesianNetwork("test")
    large_monty_network.add_nodes(s0, s1, s2, s3, s4, s5)
    large_monty_network.add_transition(s0, s1)
    large_monty_network.add_transition(s1, s3)
    large_monty_network.add_transition(s2, s3)
    large_monty_network.add_transition(s4, s5)
    large_monty_network.add_transition(s5, s2)
    large_monty_network.add_transition(s0, s2)
    large_monty_network.bake()

def setup_random_mixed():
    numpy.random.seed(0)
    global X
    X = numpy.array([
        numpy.random.choice([True, False], size=50),
        numpy.random.choice(['A', 'B'], size=50),
        numpy.random.choice(2, size=50)
    ], dtype=object).T.copy()

    global weights
    weights = numpy.abs(numpy.random.randn(50))

    global data_generator
    data_generator = DataGenerator(X, weights)

    global model
    model = BayesianNetwork.from_samples(X)

def teardown():
    pass


@with_setup(setup_monty, teardown)
def test_check_input_dict():
    obs = {'guest' : 'A'}
    _check_input(obs, monty_network)

    obs = {'guest' : 'NaN'}
    assert_raises(ValueError, _check_input, obs, monty_network)

    obs = {'guest' : None}
    assert_raises(ValueError, _check_input, obs, monty_network)

    obs = {'guest' : numpy.nan}
    assert_raises(ValueError, _check_input, obs, monty_network)

    obs = {'guest' : 'NaN', 'prize' : 'B'}
    assert_raises(ValueError, _check_input, obs, monty_network)

    obs = {'guest' : 'A', 'prize' : 'C'}
    _check_input(obs, monty_network)

    obs = {'guest' : 'A', 'prize' : 'C', 'monty' : 'C'}
    _check_input(obs, monty_network)

    obs = {'guest' : DiscreteDistribution({'A' : 0.25,
        'B' : 0.25, 'C' : 0.50})}
    _check_input(obs, monty_network)

    obs = {'hello' : 'A', 'prize' : 'B'}
    assert_raises(ValueError, _check_input, obs, monty_network)


@with_setup(setup_monty, teardown)
def test_check_input_list_of_dicts():
    obs = {'guest' : 'A'}
    _check_input([obs], monty_network)

    obs = {'guest' : 'NaN'}
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = {'guest' : None}
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = {'guest' : numpy.nan}
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = {'guest' : 'NaN', 'prize' : 'B'}
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = {'guest' : 'A', 'prize' : 'C'}
    _check_input([obs], monty_network)

    obs = {'guest' : 'A', 'prize' : 'C', 'monty' : 'C'}
    _check_input([obs], monty_network)

    obs = {'guest' : DiscreteDistribution({'A' : 0.25,
        'B' : 0.25, 'C' : 0.50})}
    _check_input([obs], monty_network)

    obs = {'hello' : 'A', 'prize' : 'B'}
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = [{'guest' : 'A'}, {'guest' : 'A', 'prize' : 'C'},
        {'guest' : 'A', 'prize' : 'C', 'monty' : 'C'},
        {'guest' : DiscreteDistribution({'A' : 0.25,
            'B' : 0.25, 'C' : 0.50})}]
    _check_input(obs, monty_network)

    obs.append({'guest' : 'NaN', 'prize' : 'B'})
    assert_raises(ValueError, _check_input, obs, monty_network)


@with_setup(setup_monty, teardown)
def test_check_input_list_of_lists():
    obs = ['A', None, None]
    _check_input([obs], monty_network)

    obs = ['A', numpy.nan, numpy.nan]
    _check_input([obs], monty_network)

    obs = numpy.array(['A', None, None])
    _check_input([obs], monty_network)

    obs = numpy.array(['A', numpy.nan, numpy.nan])
    _check_input([obs], monty_network)

    obs = numpy.array(['A', 'B', 'C'])
    _check_input([obs], monty_network)

    obs = numpy.array(['NaN', numpy.nan, numpy.nan])
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = numpy.array(['A', 'B', 'D'])
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = ['A']
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = ['A', 'C', 'E', 'F']
    assert_raises(ValueError, _check_input, [obs], monty_network)

    d = DiscreteDistribution({'A': 0.25, 'B': 0.25, 'C': 0.25})
    obs = [d, None, None]
    _check_input([obs], monty_network)

    e = DiscreteDistribution({'A': 0.25, 'B': 0.25, 'D': 0.25})
    obs = [e, None, None]
    assert_raises(ValueError, _check_input, [obs], monty_network)

    obs = [['A', None, None], ['A', numpy.nan, numpy.nan], ['A', 'B', 'C'],
        ['A', None, 'C'], [None, 'B', 'C'], [d, None, None]]
    _check_input(obs, monty_network)

    obs.append([e, None, None])
    assert_raises(ValueError, _check_input, obs, monty_network)


@with_setup(setup_titanic, teardown)
def test_titanic_network():
    assert_almost_equal(passenger.log_probability('survive'), numpy.log(0.6))
    assert_almost_equal(passenger.log_probability('survive'), numpy.log(0.6))

    assert_almost_equal(gender.log_probability(('survive', 'male')),   float("-inf"))
    assert_almost_equal(gender.log_probability(('survive', 'female')), 0.0)
    assert_almost_equal(gender.log_probability(('perish', 'male')),    0.0)
    assert_almost_equal(gender.log_probability(('perish', 'female')),  float("-inf"))

    assert_almost_equal(tclass.log_probability(('survive', 'first')), float("-inf"))
    assert_almost_equal(tclass.log_probability(('survive', 'second')), 0.0)
    assert_almost_equal(tclass.log_probability(('survive', 'third')), float("-inf"))
    assert_almost_equal(tclass.log_probability(('perish', 'first')), 0.0)
    assert_almost_equal(tclass.log_probability(('perish', 'second')), float("-inf"))
    assert_almost_equal(tclass.log_probability(('perish', 'third')), float("-inf"))


@with_setup(setup_titanic, teardown)
def test_guest_titanic():
    male = titanic_network.predict_proba({'gender': 'male'})
    female = titanic_network.predict_proba({'gender': 'female'})

    assert_equal(female[0].log_probability("survive"), 0.0)
    assert_equal(female[0].log_probability("perish"), float("-inf"))

    assert_equal(female[1], 'female')
    assert_equal(female[2].log_probability("first"), float("-inf"))
    assert_equal(female[2].log_probability("second"), 0.0)
    assert_equal(female[2].log_probability("third"), float("-inf"))

    assert_equal(male[0].log_probability("survive"), float("-inf"))
    assert_equal(male[0].log_probability("perish"), 0.0)

    assert_equal(male[1], 'male')

    assert_equal(male[2].log_probability("first"), 0.0)
    assert_equal(male[2].log_probability("second"), float("-inf"))
    assert_equal(male[2].log_probability("third"), float("-inf"))

    titanic_network2 = BayesianNetwork.from_json(titanic_network.to_json())


@with_setup(setup_large_monty, teardown)
def test_large_monty():
    assert_almost_equal(large_monty.log_probability(('A', 'A', 'C')), numpy.log(0.5))
    assert_almost_equal(large_monty.log_probability(('B', 'B', 'C')), numpy.log(0.5))
    assert_equal(large_monty.log_probability(('C', 'C', 'C')), float("-inf"))

    data = [[True,  'A', 'A', 'C', 1, True],
            [True,  'A', 'A', 'C', 0, True],
            [False, 'A', 'A', 'B', 1, False],
            [False, 'A', 'A', 'A', 2, False],
            [False, 'A', 'A', 'C', 1, False],
            [False, 'B', 'B', 'B', 2, False],
            [False, 'B', 'B', 'C', 0, False],
            [True,  'C', 'C', 'A', 2, True],
            [True,  'C', 'C', 'C', 1, False],
            [True,  'C', 'C', 'C', 0, False],
            [True,  'C', 'C', 'C', 2, True],
            [True,  'C', 'B', 'A', 1, False]]

    large_monty_network.fit(data)

    assert_almost_equal(large_monty.log_probability(('A', 'A', 'C')), numpy.log(0.6))
    assert_almost_equal(large_monty.log_probability(('B', 'B', 'C')), numpy.log(0.5))
    assert_almost_equal(large_monty.log_probability(('C', 'C', 'C')), numpy.log(0.75))


@with_setup(setup_large_monty, teardown)
def test_large_monty_friend():
    assert_almost_equal(large_monty_friend.log_probability(True), numpy.log(0.5))
    assert_almost_equal(large_monty_friend.log_probability(False), numpy.log(0.5))

    data = [[True,  'A', 'A', 'C', 1, True],
            [True,  'A', 'A', 'C', 0, True],
            [False, 'A', 'A', 'B', 1, False],
            [False, 'A', 'A', 'A', 2, False],
            [False, 'A', 'A', 'C', 1, False],
            [False, 'B', 'B', 'B', 2, False],
            [False, 'B', 'B', 'C', 0, False],
            [True,  'C', 'C', 'A', 2, True],
            [True,  'C', 'C', 'C', 1, False],
            [True,  'C', 'C', 'C', 0, False],
            [True,  'C', 'C', 'C', 2, True],
            [True,  'C', 'B', 'A', 1, False]]

    large_monty_network.fit(data)

    assert_almost_equal(large_monty_friend.log_probability(True), numpy.log(7. / 12))
    assert_almost_equal(large_monty_friend.log_probability(False), numpy.log(5. / 12))


@with_setup(setup_large_monty, teardown)
def test_large_monty_remaining():
    model = large_monty_remaining

    assert_almost_equal(model.log_probability(0), numpy.log(0.1))
    assert_almost_equal(model.log_probability(1), numpy.log(0.7))
    assert_almost_equal(model.log_probability(2), numpy.log(0.2))

    data = [[True,  'A', 'A', 'C', 1, True],
            [True,  'A', 'A', 'C', 0, True],
            [False, 'A', 'A', 'B', 1, False],
            [False, 'A', 'A', 'A', 2, False],
            [False, 'A', 'A', 'C', 1, False],
            [False, 'B', 'B', 'B', 2, False],
            [False, 'B', 'B', 'C', 0, False],
            [True,  'C', 'C', 'A', 2, True],
            [True,  'C', 'C', 'C', 1, False],
            [True,  'C', 'C', 'C', 0, False],
            [True,  'C', 'C', 'C', 2, True],
            [True,  'C', 'B', 'A', 1, False]]

    large_monty_network.fit(data)

    assert_almost_equal(model.log_probability(0), numpy.log(3. / 12))
    assert_almost_equal(model.log_probability(1), numpy.log(5. / 12))
    assert_almost_equal(model.log_probability(2), numpy.log(4. / 12))

@with_setup(setup_large_monty, teardown)
def test_large_monty_network_log_probability():
    model = large_monty_network

    data = numpy.array([[True,  'A', 'A', 'C', 1, True],
            [True,  'A', 'A', 'C', 0, True],
            [False, 'A', 'A', 'B', 1, False],
            [False, 'A', 'A', 'A', 2, False],
            [False, 'A', 'A', 'C', 1, False],
            [False, 'B', 'B', 'B', 2, False],
            [False, 'B', 'B', 'C', 0, False],
            [True,  'C', 'C', 'A', 2, True],
            [True,  'C', 'C', 'C', 1, False],
            [True,  'C', 'C', 'C', 0, False],
            [True,  'C', 'C', 'C', 2, True],
            [True,  'C', 'B', 'A', 1, False]],
            dtype=object)

    logp = [-3.863233, -8.581732, float("-inf"), float("-inf"), float("-inf"),
        float("-inf"), -5.013138, -6.279147, float("-inf"), float("-inf"),
        float("-inf"), -4.150915]
    logp1 = model.log_probability(data)

    assert_array_almost_equal(logp, logp1)

    model.fit(data)

    logp = [-5.480639, -5.480639, -4.60517 , -5.298317, -3.506558, -5.192957,
       -5.192957, -4.74667 , -2.667228, -3.360375, -3.648057, -3.072693]
    logp2 = model.log_probability(data)

    assert_array_almost_equal(logp2, logp)

@with_setup(setup_large_monty, teardown)
def test_large_monty_network_log_probability_parallel():
    model = large_monty_network

    data = numpy.array([[True,  'A', 'A', 'C', 1, True],
            [True,  'A', 'A', 'C', 0, True],
            [False, 'A', 'A', 'B', 1, False],
            [False, 'A', 'A', 'A', 2, False],
            [False, 'A', 'A', 'C', 1, False],
            [False, 'B', 'B', 'B', 2, False],
            [False, 'B', 'B', 'C', 0, False],
            [True,  'C', 'C', 'A', 2, True],
            [True,  'C', 'C', 'C', 1, False],
            [True,  'C', 'C', 'C', 0, False],
            [True,  'C', 'C', 'C', 2, True],
            [True,  'C', 'B', 'A', 1, False]], dtype=object)

    logp = [-3.863233, -8.581732, float("-inf"), float("-inf"), float("-inf"),
        float("-inf"), -5.013138, -6.279147, float("-inf"), float("-inf"),
        float("-inf"), -4.150915]

    logp1 = model.log_probability(data, n_jobs=2)

    assert_array_almost_equal(logp, logp1)

    model.fit(data)

    logp = [-5.480639, -5.480639, -4.60517 , -5.298317, -3.506558, -5.192957,
       -5.192957, -4.74667 , -2.667228, -3.360375, -3.648057, -3.072693]
    logp2 = model.log_probability(data, n_jobs=2)

    assert_array_almost_equal(logp2, logp)

@with_setup(setup_large_monty, teardown)
def test_large_monty_prize():
    assert_almost_equal(large_monty_prize.log_probability(
        (True,  True,  'A')), numpy.log(0.3))
    assert_almost_equal(large_monty_prize.log_probability(
        (True,  False, 'C')), numpy.log(0.4))
    assert_almost_equal(large_monty_prize.log_probability(
        (False, True,  'B')), numpy.log(0.9))
    assert_almost_equal(large_monty_prize.log_probability(
        (False, False, 'A')), float("-inf"))

    data = [[True,  'A', 'A', 'C', 1, True],
            [True,  'A', 'A', 'C', 0, True],
            [False, 'A', 'A', 'B', 1, False],
            [False, 'A', 'A', 'A', 2, False],
            [False, 'A', 'A', 'C', 1, False],
            [False, 'B', 'B', 'B', 2, False],
            [False, 'B', 'B', 'C', 0, False],
            [True,  'C', 'C', 'A', 2, True],
            [True,  'C', 'C', 'C', 1, False],
            [True,  'C', 'C', 'C', 0, False],
            [True,  'C', 'C', 'C', 2, True],
            [True,  'C', 'B', 'A', 1, False]]

    large_monty_network.fit(data)

    assert_almost_equal(large_monty_prize.log_probability(
        (True, True, 'C')), numpy.log(0.5))
    assert_equal(large_monty_prize.log_probability(
        (True, True, 'B')), float("-inf"))

    a = large_monty_prize.log_probability((True, False, 'A'))
    b = large_monty_prize.log_probability((True, False, 'B'))
    c = large_monty_prize.log_probability((True, False, 'C'))

    assert_almost_equal(a, b)
    assert_almost_equal(b, c)

    assert_equal(large_monty_prize.log_probability(
        (False, False, 'C')), float("-inf"))
    assert_almost_equal(large_monty_prize.log_probability(
        (False, True, 'C')), numpy.log(2. / 3))


def assert_discrete_equal(x, y, z=8):
    xd, yd = x.parameters[0], y.parameters[0]
    for key, value in xd.items():
        if round(yd[key], z) != round(value, z):
            raise ValueError("{} != {}".format(yd[key], value))


@with_setup(setup_monty, teardown)
def test_guest_monty():
    a = monty_network.predict_proba({'guest': 'A'})
    b = monty_network.predict_proba({'guest': 'B'})
    c = monty_network.predict_proba({'guest': 'C'})

    prize_correct = DiscreteDistribution(
        {'A': 1. / 3, 'B': 1. / 3, 'C': 1. / 3})

    assert_discrete_equal(a[prize_index], b[prize_index])
    assert_discrete_equal(a[prize_index], c[prize_index])
    assert_discrete_equal(a[prize_index], prize_correct)

    assert_discrete_equal(a[monty_index], DiscreteDistribution(
        {'A': 0.0, 'B': 1. / 2, 'C': 1. / 2}))
    assert_discrete_equal(b[monty_index], DiscreteDistribution(
        {'A': 1. / 2, 'B': 0.0, 'C': 1. / 2}))
    assert_discrete_equal(c[monty_index], DiscreteDistribution(
        {'A': 1. / 2, 'B': 1. / 2, 'C': 0.0}))


@with_setup(setup_monty, teardown)
def test_guest_with_monty():
    b = monty_network.predict_proba({'guest': 'A', 'monty': 'B'})
    c = monty_network.predict_proba({'guest': 'A', 'monty': 'C'})

    assert_equal(b[guest_index], 'A')
    assert_equal(b[monty_index], 'B')
    assert_discrete_equal(b[prize_index], DiscreteDistribution(
        {'A': 1. / 3, 'B': 0.0, 'C': 2. / 3}))

    assert_equal(c[guest_index], 'A')
    assert_equal(c[monty_index], 'C')
    assert_discrete_equal(c[prize_index], DiscreteDistribution(
        {'A': 1. / 3, 'B': 2. / 3, 'C': 0.0}))


@with_setup(setup_monty, teardown)
def test_monty():
    a = monty_network.predict_proba({'monty': 'A'})

    assert_equal(a[monty_index], 'A')
    assert_discrete_equal(a[guest_index], a[prize_index])
    assert_discrete_equal(a[guest_index], DiscreteDistribution(
        {'A': 0.0, 'B': 1. / 2, 'C': 1. / 2}))


@with_setup(setup_monty, teardown)
def test_predict():
    obs = [['A', None, 'B'],
           ['A', None, 'C'],
           ['A', 'B', 'C']]

    predictions = monty_network.predict(obs)

    assert_array_equal(predictions,
                       [
                         ['A', 'C', 'B'],
                         ['A', 'B', 'C'],
                         ['A', 'B', 'C']
                       ])

    assert_array_equal(obs,
                       [
                         ['A', None, 'B'],
                         ['A', None, 'C'],
                         ['A', 'B', 'C']
                       ])

@with_setup(setup_monty, teardown)
def test_rejection_sampling():
    numpy.random.seed(0)
    predictions = monty_network._rejection(n=10,evidences=[{'guest':'A', 'monty':'B'}])
    (unique, counts) = numpy.unique(predictions[:,1], return_counts=True)
    assert_array_equal(unique, ['A', 'C'])
    assert counts[0] > 0 and counts[0] < 4
    # Need to find where random seed is changed so next test can work 
    # assert_array_equal(predictions,
    #                    [['A', 'C', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'A', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'C', 'B'],
    #                     ['A', 'C', 'B']])

@with_setup(setup_monty, teardown)
def test_gibbs_sampling():
    # evidences = [['A', None, 'B'],
    #              ['A', None, 'C'],
    #              ['A', 'B', 'C' ]]
    predictions = monty_network._gibbs(n=1000,evidences=[{'guest': 'A', 'monty': 'B'}])
    values, counts = numpy.unique(predictions[:, 1], return_counts=True)
    # will fail from time to time
    # need to fix the seed
    assert(abs(counts[0]-340) < 15)

@with_setup(setup_monty, teardown)
def test_predict_parallel():
    obs = [['A', None, 'B'],
           ['A', None, 'C'],
           ['A', 'B', 'C']]

    predictions = monty_network.predict(obs, n_jobs=2)
    assert_array_equal(predictions,
                       [
                         ['A', 'C', 'B'],
                         ['A', 'B', 'C'],
                         ['A', 'B', 'C']
                       ])

    assert_array_equal(obs,
                       [
                         ['A', None, 'B'],
                         ['A', None, 'C'],
                         ['A', 'B', 'C']
                       ])

@with_setup(setup_monty, teardown)
def test_predict_datagenerator():
    obs = [['A', None, 'B'],
           ['A', None, 'C'],
           ['A', 'B', 'C']]

    X = DataGenerator(obs)

    predictions = monty_network.predict(X)

    assert_array_equal(predictions,
                       [
                         ['A', 'C', 'B'],
                         ['A', 'B', 'C'],
                         ['A', 'B', 'C']
                       ])

    assert_array_equal(obs,
                       [
                         ['A', None, 'B'],
                         ['A', None, 'C'],
                         ['A', 'B', 'C']
                       ])

@with_setup(setup_monty, teardown)
def test_numpy_predict():
    obs = numpy.array([['A', None, 'B'],
                    ['A', None, 'C'],
                    ['A', 'B', 'C']])

    predictions = monty_network.predict(obs)

    assert_array_equal(predictions,
                       [
                         ['A', 'C', 'B'],
                         ['A', 'B', 'C'],
                         ['A', 'B', 'C']
                       ])

    assert_array_equal(obs,
                       [
                         ['A', None, 'B'],
                         ['A', None, 'C'],
                         ['A', 'B', 'C']
                       ])

@with_setup(setup_monty, teardown)
def test_numpy_predict_parallel():
    obs = numpy.array([['A', None, 'B'],
                    ['A', None, 'C'],
                    ['A', 'B', 'C']])

    predictions = monty_network.predict(obs, n_jobs=2)

    assert_array_equal(predictions,
                       [
                         ['A', 'C', 'B'],
                         ['A', 'B', 'C'],
                         ['A', 'B', 'C']
                       ])

    assert_array_equal(obs,
                       [
                         ['A', None, 'B'],
                         ['A', None, 'C'],
                         ['A', 'B', 'C']
                       ])

@with_setup(setup_monty, teardown)
def test_numpy_predict_datagenerator():
    obs = numpy.array([['A', None, 'B'],
                    ['A', None, 'C'],
                    ['A', 'B', 'C']])

    X = DataGenerator(obs)

    predictions = monty_network.predict(X)

    assert_array_equal(predictions,
                       [
                         ['A', 'C', 'B'],
                         ['A', 'B', 'C'],
                         ['A', 'B', 'C']
                       ])

    assert_array_equal(obs,
                       [
                         ['A', None, 'B'],
                         ['A', None, 'C'],
                         ['A', 'B', 'C']
                       ])


@with_setup(setup_monty, teardown)
def test_single_dict_predict_proba():
    obs = {'guest': 'A',  'monty': 'B'}
    y = DiscreteDistribution({'A': 1./3, 'B': 0., 'C': 2./3})
    y_hat = monty_network.predict_proba(obs)

    assert_equal(y_hat[0], 'A')
    assert_equal(y_hat[2], 'B')
    assert_discrete_equal(y_hat[1], y)


@with_setup(setup_large_monty, teardown)
def test_single_dict_large_predict_proba():
    obs = {'large_monty_friend' : True,  'large_monty_guest': 'A',
        'large_monty_prize': 'A', 'large_monty': 'C'}
    y1 = DiscreteDistribution({0: 0.0472, 1: 0.781, 2: 0.17167})
    y2 = DiscreteDistribution({True: 0.8562, False: 0.143776})
    y_hat = large_monty_network.predict_proba(obs)

    assert_equal(y_hat[0], True)
    assert_equal(y_hat[1], 'A')
    assert_equal(y_hat[2], 'A')
    assert_equal(y_hat[3], 'C')
    assert_discrete_equal(y_hat[4], y1, 3)
    assert_discrete_equal(y_hat[5], y2, 3)

    obs = {'large_monty_friend' : True, 'large_monty_prize': 'A',
        'large_monty': 'C', 'large_monty_remaining' : 2}
    y1 = DiscreteDistribution({'A': 0.5, 'B': 0.5, 'C': 0.0})
    y2 = DiscreteDistribution({True: 0.75, False: 0.25})
    y_hat = large_monty_network.predict_proba(obs)

    assert_equal(y_hat[0], True)
    assert_equal(y_hat[2], 'A')
    assert_equal(y_hat[3], 'C')
    assert_equal(y_hat[4], 2)
    assert_discrete_equal(y_hat[1], y1)
    assert_discrete_equal(y_hat[5], y2)


@with_setup(setup_monty, teardown)
def test_list_of_lists_predict_proba():
    obs = [['A', None, 'B']]
    y = DiscreteDistribution({'A': 1./3, 'B': 0., 'C': 2./3})
    y_hat = monty_network.predict_proba(obs)

    assert_equal(y_hat[0][0], 'A')
    assert_equal(y_hat[0][2], 'B')
    assert_discrete_equal(y_hat[0][1], y)


@with_setup(setup_large_monty, teardown)
def test_list_of_lists_large_predict_proba():
    obs = [[True,  'A', 'A', 'C', None, None]]
    y1 = DiscreteDistribution({0: 0.0472, 1: 0.781, 2: 0.17167})
    y2 = DiscreteDistribution({True: 0.8562, False: 0.143776})
    y_hat = large_monty_network.predict_proba(obs)

    assert_equal(y_hat[0][0], True)
    assert_equal(y_hat[0][1], 'A')
    assert_equal(y_hat[0][2], 'A')
    assert_equal(y_hat[0][3], 'C')
    assert_discrete_equal(y_hat[0][4], y1, 3)
    assert_discrete_equal(y_hat[0][5], y2, 3)

    obs = [[True, None, 'A', 'C', 2, None]]
    y1 = DiscreteDistribution({'A': 0.5, 'B': 0.5, 'C': 0.0})
    y2 = DiscreteDistribution({True: 0.75, False: 0.25})
    y_hat = large_monty_network.predict_proba(obs)

    assert_equal(y_hat[0][0], True)
    assert_equal(y_hat[0][2], 'A')
    assert_equal(y_hat[0][3], 'C')
    assert_equal(y_hat[0][4], 2)
    assert_discrete_equal(y_hat[0][1], y1)
    assert_discrete_equal(y_hat[0][5], y2)


@with_setup(setup_monty, teardown)
def test_list_of_dicts_predict_proba():
    obs = [{'guest': 'A',  'monty': 'B'}]
    y = DiscreteDistribution({'A': 1./3, 'B': 0., 'C': 2./3})
    y_hat = monty_network.predict_proba(obs)

    assert_equal(y_hat[0][0], 'A')
    assert_equal(y_hat[0][2], 'B')
    assert_discrete_equal(y_hat[0][1], y)


@with_setup(setup_large_monty, teardown)
def test_list_of_dicts_large_predict_proba():
    obs = [{'large_monty_friend' : True,  'large_monty_guest': 'A',
        'large_monty_prize': 'A', 'large_monty': 'C'}]
    y1 = DiscreteDistribution({0: 0.0472, 1: 0.781, 2: 0.17167})
    y2 = DiscreteDistribution({True: 0.8562, False: 0.143776})
    y_hat = large_monty_network.predict_proba(obs)

    assert_equal(y_hat[0][0], True)
    assert_equal(y_hat[0][1], 'A')
    assert_equal(y_hat[0][2], 'A')
    assert_equal(y_hat[0][3], 'C')
    assert_discrete_equal(y_hat[0][4], y1, 3)
    assert_discrete_equal(y_hat[0][5], y2, 3)

    obs = [{'large_monty_friend' : True, 'large_monty_prize': 'A',
        'large_monty': 'C', 'large_monty_remaining' : 2}]
    y1 = DiscreteDistribution({'A': 0.5, 'B': 0.5, 'C': 0.0})
    y2 = DiscreteDistribution({True: 0.75, False: 0.25})
    y_hat = large_monty_network.predict_proba(obs)

    assert_equal(y_hat[0][0], True)
    assert_equal(y_hat[0][2], 'A')
    assert_equal(y_hat[0][3], 'C')
    assert_equal(y_hat[0][4], 2)
    assert_discrete_equal(y_hat[0][1], y1)
    assert_discrete_equal(y_hat[0][5], y2)


@with_setup(setup_monty, teardown)
def test_list_of_dicts_predict_proba_parallel():
    obs = [{'guest': 'A',  'monty': 'B'},
           {'guest': 'B', 'prize': 'A'},
           {'monty': 'C', 'prize': 'B'},
           {'monty': 'B'}, {'prize': 'A'}]
    y = DiscreteDistribution({'A': 1./3, 'B': 0., 'C': 2./3})
    y_hat = monty_network.predict_proba(obs, n_jobs=2)

    assert_equal(y_hat[0][0], 'A')
    assert_equal(y_hat[0][2], 'B')
    assert_discrete_equal(y_hat[0][1], y)

    assert_equal(y_hat[1][0], 'B')
    assert_equal(y_hat[1][1], 'A')

    assert_equal(y_hat[3][2], 'B')
    assert_equal(y_hat[4][1], 'A')


@with_setup(setup_monty, teardown)
def test_raise_error():
    obs = [['green', 'cat', None]]
    assert_raises(ValueError, monty_network.predict, obs)

    obs = [['A', 'b', None]]
    assert_raises(ValueError, monty_network.predict, obs)

    obs = [['none', 'B', None]]
    assert_raises(ValueError, monty_network.predict, obs)

    obs = [['NaN', 'B', None]]
    assert_raises(ValueError, monty_network.predict, obs)

    obs = [['A', 'C', 'D']]
    assert_raises(ValueError, monty_network.predict, obs)


def test_exact_structure_learning():
    logps = -19.8282, -345.9527, -4847.59688, -604.0190
    for X, logp in zip(datasets, logps):
        model = BayesianNetwork.from_samples(X, algorithm='exact')
        model2 = BayesianNetwork.from_samples(X, algorithm='exact-dp')
        assert_almost_equal(model.log_probability(X).sum(), model2.log_probability(X).sum())
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)

def test_exact_penalized_structure_learning():
    n_parents = [(5, 3, 4), (10, 0, 1), (21, 0, 8), (26, 3, 21)]
    for X, n_parents in zip(datasets, n_parents):
        model = BayesianNetwork.from_samples(X, algorithm='exact', penalty=0)
        model2 = BayesianNetwork.from_samples(X, algorithm='exact-dp', penalty=0)
        assert_equal(sum(map(len, model.structure)), n_parents[0])
        assert_equal(sum(map(len, model2.structure)), n_parents[0])

        model = BayesianNetwork.from_samples(X, algorithm='exact')
        model2 = BayesianNetwork.from_samples(X, algorithm='exact-dp')
        assert_equal(sum(map(len, model.structure)), n_parents[1])
        assert_equal(sum(map(len, model2.structure)), n_parents[1])

        model = BayesianNetwork.from_samples(X, algorithm='exact', penalty=1)
        model2 = BayesianNetwork.from_samples(X, algorithm='exact-dp', penalty=1)
        assert_equal(sum(map(len, model.structure)), n_parents[2])
        assert_equal(sum(map(len, model2.structure)), n_parents[2])

        model = BayesianNetwork.from_samples(X, algorithm='exact', penalty=100)
        model2 = BayesianNetwork.from_samples(X, algorithm='exact-dp', penalty=100)
        assert_equal(sum(map(len, model.structure)), 0)
        assert_equal(sum(map(len, model2.structure)), 0)

def test_exact_structure_learning_include_edges():
    for X in datasets:
        model = BayesianNetwork.from_samples(X, algorithm='exact', 
            include_edges=[(1, 3)])
        assert_equal(model.structure[3], (1,))

        model = BayesianNetwork.from_samples(X, algorithm='exact')
        assert_not_equal(model.structure[3], (1,))

def test_exact_dp_structure_learning_include_edges():
    for X in datasets:
        model = BayesianNetwork.from_samples(X, algorithm='exact-dp', 
            include_edges=[(1, 3)])
        assert_equal(model.structure[3], (1,))

        model = BayesianNetwork.from_samples(X, algorithm='exact-dp')
        assert_not_equal(model.structure[3], (1,))

def test_exact_structure_learning_exclude_edges():
    for X in datasets:
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]

        # Learn constrained network
        model = BayesianNetwork.from_samples(X, algorithm='exact', 
            exclude_edges=[(1, d-1), (d-1, d-2)])    
        assert_not_equal(model.structure[-1], (1,))
        assert_not_equal(model.structure[-2], (d-1,))
        assert_equal(model.structure[-2], (1,))

def test_exact_dp_structure_learning_exclude_edges():
    for X in datasets:
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]

        # Learn constrained network
        model = BayesianNetwork.from_samples(X, algorithm='exact-dp', 
            exclude_edges=[(1, d-1), (d-1, d-2)])    
        assert_not_equal(model.structure[-1], (1,))
        assert_not_equal(model.structure[-2], (d-1,))
        assert_equal(model.structure[-2], (1,))

def test_constrained_sl_structure_learning_exclude_edges():
    for X in datasets:
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]

        cg = DiGraph()
        n = tuple(range(d))
        cg.add_edge(n, n)

        # Learn constrained network
        model = BayesianNetwork.from_samples(X, algorithm='greedy', 
            constraint_graph=cg, exclude_edges=[(1, d-1), (d-1, d-2)])    
        assert_not_equal(model.structure[-1], (1,))
        assert_not_equal(model.structure[-2], (d-1,))
        assert_equal(model.structure[-2], (1,))

def test_constrained_parents_structure_learning_exclude_edges():
    for X in datasets:
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]
        d1 = int(numpy.ceil(d / 2))
        
        # Node groups
        g1 = tuple(range(0, d1))
        g2 = tuple(range(d1, d))
        
        # Constraint graph:
        cg = DiGraph()
        cg.add_edge(g1, g2)

        # Learn constrained network
        model1 = BayesianNetwork.from_samples(X, algorithm='exact', 
            constraint_graph=cg, exclude_edges=[(1, d-1)])
        assert_not_equal(model1.structure[-1], (1,))
        assert_equal(model1.structure[-2], (1,))

        model2 = BayesianNetwork.from_samples(X, algorithm='exact',
            constraint_graph=cg)
        assert_equal(model2.structure[-1], (1,))
        assert_equal(model2.structure[-2], (1,))

    X = numpy.random.randint(2, size=(50, 8))
    X[:,0] = X[:,4]
    X[:,1] = X[:,7]
    X[:,2] = X[:,7]

    cg = DiGraph()
    n1 = (0, 2, 3, 5, 6)
    n2 = (1, 4, 7)
    cg.add_edge(n1, n2)

    model = BayesianNetwork.from_samples(X, algorithm='exact', 
        constraint_graph=cg, exclude_edges=[(0, 4), (2, 7)])
    assert_not_equal(model.structure[7], (2,))
    assert_not_equal(model.structure[4], (0,))

    model = BayesianNetwork.from_samples(X, algorithm='exact',
        constraint_graph=cg)
    assert_equal(model.structure[7], (2,))
    assert_equal(model.structure[4], (0,))

def test_constrained_slap_structure_learning_exclude_edges():
    for X in datasets:
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]
        d1 = int(numpy.ceil(d / 2))
        
        # Node groups
        g1 = tuple(range(0, d1))
        g2 = tuple(range(d1, d))
        
        # Constraint graph:
        cg = DiGraph()
        cg.add_edge(g1, g2)
        cg.add_edge(g2, g2)

        # Learn constrained network
        model1 = BayesianNetwork.from_samples(X, algorithm='exact', 
            constraint_graph=cg, exclude_edges=[(1, d-1)])
        assert_not_equal(model1.structure[-1], (1,))
        assert_equal(model1.structure[-1], (d-2,))

        model2 = BayesianNetwork.from_samples(X, algorithm='exact',
            constraint_graph=cg)
        assert_equal(model2.structure[-1], (d-2,))

    X = numpy.random.randint(2, size=(50, 8))
    X[:,0] = X[:,4]
    X[:,1] = X[:,7]
    X[:,2] = X[:,7]

    cg = DiGraph()
    n1 = (0, 2, 3, 5, 6)
    n2 = (1, 4, 7)
    cg.add_edge(n1, n2)
    cg.add_edge(n2, n2)

    model = BayesianNetwork.from_samples(X, algorithm='exact', 
        constraint_graph=cg, exclude_edges=[(0, 4), (2, 7)])
    assert_not_equal(model.structure[7], (2,))
    assert_not_equal(model.structure[4], (0,))

    model = BayesianNetwork.from_samples(X, algorithm='exact',
        constraint_graph=cg)
    assert_equal(model.structure[7], (2,))
    assert_equal(model.structure[4], (0,))

def test_constrained_parents_structure_learning():
    logps1 = [-12.2173, -207.3633, -3462.7469, -480.0970]
    logps2 = [-10.8890, -207.3633, -3462.7469, -480.0970]

    for X, logp1, logp2 in zip(datasets, logps1, logps2):
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]
        d1 = int(numpy.ceil(d / 2))
        
        # Node groups
        g1 = tuple(range(0, d1))
        g2 = tuple(range(d1, d))
        
        # Constraint graph:
        cg = DiGraph()
        cg.add_edge(g1, g2)

        # Learn constrained network
        model1 = BayesianNetwork.from_samples(X, algorithm='exact', 
            constraint_graph=cg)
        assert_almost_equal(model1.log_probability(X).sum(), logp1, 4)
        
        # Check structure constraints satisfied
        for node in g1:
            assert_equal(0, len(model1.structure[node]))
        
        assert_equal(model1.structure[-1], (1,))
        assert_equal(model1.structure[-2], (1,))

        model2 = BayesianNetwork.from_samples(X, algorithm='exact')
        assert_almost_equal(model2.log_probability(X).sum(), logp2, 4)
        assert_equal(model2.structure[-1], (d-2,))
        assert_equal(model2.structure[-2], (1,))

def test_constrained_slap_structure_learning():
    logps = [-21.7780, -345.9527, -4847.5969, -611.0356]

    for X, logp in zip(datasets, logps):
        d = X.shape[1]
        d1 = int(numpy.ceil(d / 2))
        
        # Node groups
        g1 = tuple(range(0, d1))
        g2 = tuple(range(d1, d))
        
        # Constraint graph:
        cg = DiGraph()
        cg.add_edge(g1, g2)
        cg.add_edge(g2, g2)

        # Learn constrained network
        model = BayesianNetwork.from_samples(X, algorithm='exact', constraint_graph=cg)
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)
        
        # Check structure constraints satisfied
        for node in g1:
            assert_equal(0, len(model.structure[node]))

def test_from_structure():
    X = datasets[1]
    structure = ((1, 2), (4,), (), (), (3,))
    model = BayesianNetwork.from_structure(X, structure=structure)

    assert_equal(model.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model2 = BayesianNetwork.from_json(model.to_json())
    assert_equal(model2.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model_dtype = type(model.states[0].distribution.parameters[0][0][0])
    model2_dtype = type(model2.states[0].distribution.parameters[0][0][0])
    assert_equal(model_dtype, model2_dtype)

def test_robust_from_structure():
    X = datasets[1]
    structure = ((1, 2), (4,), (), (), (3,))
    model = BayesianNetwork.from_structure(X, structure=structure)

    assert_equal(model.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model2 = from_json(model.to_json())
    assert_equal(model2.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model_dtype = type(model.states[0].distribution.parameters[0][0][0])
    model2_dtype = type(model2.states[0].distribution.parameters[0][0][0])
    assert_equal(model_dtype, model2_dtype)

@with_setup(setup_random_mixed)
def test_from_json():
    model2 = BayesianNetwork.from_json(model.to_json())

    logp1 = model.log_probability(X)
    logp2 = model2.log_probability(X)
    logp = [-2.304186, -1.898721, -1.898721, -2.224144, -1.898721, -1.978764,
        -1.898721, -1.898721, -1.898721, -1.898721, -1.818679, -2.384229,
        -2.304186, -1.978764, -2.304186, -2.384229, -2.304186, -2.384229,
        -2.304186, -1.978764, -2.224144, -1.818679, -1.898721, -2.304186,
        -2.304186, -1.898721, -1.818679, -1.898721, -1.818679, -2.304186,
        -1.978764, -2.224144, -1.898721, -2.304186, -1.898721, -1.818679,
        -2.304186, -1.898721, -1.898721, -2.384229, -2.224144, -1.818679,
        -2.384229, -1.978764, -1.818679, -1.978764, -1.898721, -1.818679,
        -2.224144, -1.898721]

    assert_array_almost_equal(logp1, logp2)
    assert_array_almost_equal(logp1, logp)
    assert_array_almost_equal(logp2, logp)

    model_dtype = type(list(model.states[0].distribution.parameters[0].keys())[0])
    model2_dtype = type(list(model2.states[0].distribution.parameters[0].keys())[0])
    assert_equal(model_dtype, model2_dtype)

@with_setup(setup_random_mixed)
def test_robust_from_json():
    model2 = from_json(model.to_json())

    logp1 = model.log_probability(X)
    logp2 = model2.log_probability(X)
    logp = [-2.304186, -1.898721, -1.898721, -2.224144, -1.898721, -1.978764,
        -1.898721, -1.898721, -1.898721, -1.898721, -1.818679, -2.384229,
        -2.304186, -1.978764, -2.304186, -2.384229, -2.304186, -2.384229,
        -2.304186, -1.978764, -2.224144, -1.818679, -1.898721, -2.304186,
        -2.304186, -1.898721, -1.818679, -1.898721, -1.818679, -2.304186,
        -1.978764, -2.224144, -1.898721, -2.304186, -1.898721, -1.818679,
        -2.304186, -1.898721, -1.898721, -2.384229, -2.224144, -1.818679,
        -2.384229, -1.978764, -1.818679, -1.978764, -1.898721, -1.818679,
        -2.224144, -1.898721]

    assert_array_almost_equal(logp1, logp2)
    assert_array_almost_equal(logp1, logp)
    assert_array_almost_equal(logp2, logp)

    model_dtype = type(list(model.states[0].distribution.parameters[0].keys())[0])
    model2_dtype = type(list(model2.states[0].distribution.parameters[0].keys())[0])
    assert_equal(model_dtype, model2_dtype)

def test_float64_from_json():
    X = datasets[1].astype('float64')
    structure = ((1, 2), (4,), (), (), (3,))
    model = BayesianNetwork.from_structure(X, structure=structure)

    assert_equal(model.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model2 = BayesianNetwork.from_json(model.to_json())
    assert_equal(model2.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model_dtype = type(model.states[0].distribution.parameters[0][0][0])
    model2_dtype = type(model2.states[0].distribution.parameters[0][0][0])
    assert_equal(model_dtype, model2_dtype)

def test_robust_float64_from_json():
    X = datasets[1].astype('float64')
    structure = ((1, 2), (4,), (), (), (3,))
    model = BayesianNetwork.from_structure(X, structure=structure)

    assert_equal(model.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model2 = from_json(model.to_json())
    assert_equal(model2.structure, structure)
    assert_almost_equal(model.log_probability(X).sum(), -344.38287, 4)

    model_dtype = type(model.states[0].distribution.parameters[0][0][0])
    model2_dtype = type(model2.states[0].distribution.parameters[0][0][0])
    assert_equal(model_dtype, model2_dtype)

def test_parallel_structure_learning():
    logps = -19.8282, -345.9527, -4847.59688, -604.0190
    for X, logp in zip(datasets, logps):
        model = BayesianNetwork.from_samples(X, algorithm='exact')
        model2 = BayesianNetwork.from_samples(X, algorithm='exact', n_jobs=2)
        assert_equal(model.log_probability(X).sum(), model2.log_probability(X).sum())
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)


def test_greedy_structure_learning():
    logps = -19.8282, -345.9527, -4847.59688, -611.0356
    for X, logp in zip(datasets, logps):
        model = BayesianNetwork.from_samples(X, algorithm='greedy')
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)

def test_greedy_structure_learning_include_edges():
    for X in datasets:
        model = BayesianNetwork.from_samples(X, algorithm='greedy', 
            include_edges=[(1, 3)])
        assert_equal(model.structure[3], (1,))

        model = BayesianNetwork.from_samples(X, algorithm='greedy')
        assert_not_equal(model.structure[3], (1,))

def test_greedy_structure_learning_exclude_edges():
    for X in datasets:
        X = X.copy()
        X[:,1] = X[:,-1]
        X[:,-2] = X[:,-1]

        d = X.shape[1]

        # Learn constrained network
        model = BayesianNetwork.from_samples(X, algorithm='greedy', 
            exclude_edges=[(1, d-1), (d-1, d-2)])    
        assert_not_equal(model.structure[-1], (1,))
        assert_not_equal(model.structure[-2], (d-1,))
        assert_equal(model.structure[-2], (1,))

def test_greedy_penalized_structure_learning():
    n_parents = [(5, 3, 4), (10, 0, 1), (21, 0, 5), (26, 1, 21)]
    for X, n_parents in zip(datasets, n_parents):
        model = BayesianNetwork.from_samples(X, algorithm='greedy', penalty=0)
        assert_equal(sum(map(len, model.structure)), n_parents[0])

        model = BayesianNetwork.from_samples(X, algorithm='greedy')
        assert_equal(sum(map(len, model.structure)), n_parents[1])

        model = BayesianNetwork.from_samples(X, algorithm='greedy', penalty=1)
        assert_equal(sum(map(len, model.structure)), n_parents[2])

        model = BayesianNetwork.from_samples(X, algorithm='greedy', penalty=100)
        assert_equal(sum(map(len, model.structure)), 0)

def test_chow_liu_structure_learning():
    logps = -19.8282, -344.248785, -4842.40158, -603.2370
    for X, logp in zip(datasets, logps):
        model = BayesianNetwork.from_samples(X, algorithm='chow-liu')
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)


def test_exact_nan_structure_learning():
    logps = -6.13764, -159.6505, -2055.76364, -201.73615
    for X, logp in zip(datasets_nan, logps):
        model = BayesianNetwork.from_samples(X, algorithm='exact')
        model2 = BayesianNetwork.from_samples(X, algorithm='exact-dp')

        assert_equal(model.log_probability(X).sum(), model2.log_probability(X).sum())
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)


def test_greedy_nan_structure_learning():
    logps = -7.5239, -159.6505, -2058.5706, -203.7662
    for X, logp in zip(datasets_nan, logps):
        model = BayesianNetwork.from_samples(X, algorithm='greedy')
        assert_almost_equal(model.log_probability(X).sum(), logp, 4)

@with_setup(setup_random_mixed, teardown)
def test_io_log_probability():
    X2 = DataGenerator(X)
    X3 = DataFrameGenerator(pandas.DataFrame(X))

    logp1 = model.log_probability(X)
    logp2 = model.log_probability(X2)
    logp3 = model.log_probability(X3)

    assert_array_almost_equal(logp1, logp2)
    assert_array_almost_equal(logp1, logp3)

@with_setup(setup_random_mixed, teardown)
def test_io_predict():
    X2 = DataGenerator(X)
    X3 = DataFrameGenerator(pandas.DataFrame(X))

    y_hat1 = model.predict(X)
    y_hat2 = model.predict(X2)
    y_hat3 = model.predict(X3)

    assert_array_equal(y_hat1, y_hat2)
    assert_array_equal(y_hat1, y_hat3)

@with_setup(setup_random_mixed, teardown)
def test_io_fit():
    d1 = DiscreteDistribution({True: 0.6, False: 0.4})
    d2 = ConditionalProbabilityTable([
        [True, 'A', 0.2],
        [True, 'B', 0.8],
        [False, 'A', 0.3],
        [False, 'B', 0.7]], [d1])
    d3 = ConditionalProbabilityTable([
        ['A', 0, 0.3],
        ['A', 1, 0.7],
        ['B', 0, 0.8],
        ['B', 1, 0.2]], [d2])

    n1 = Node(d1)
    n2 = Node(d2)
    n3 = Node(d3)

    model1 = BayesianNetwork()
    model1.add_nodes(n1, n2, n3)
    model1.add_edge(n1, n2)
    model1.add_edge(n2, n3)
    model1.bake()
    model1.fit(X, weights=weights)

    d1 = DiscreteDistribution({True: 0.2, False: 0.8})
    d2 = ConditionalProbabilityTable([
        [True, 'A', 0.7],
        [True, 'B', 0.2],
        [False, 'A', 0.4],
        [False, 'B', 0.6]], [d1])
    d3 = ConditionalProbabilityTable([
        ['A', 0, 0.9],
        ['A', 1, 0.1],
        ['B', 0, 0.0],
        ['B', 1, 1.0]], [d2])

    n1 = Node(d1)
    n2 = Node(d2)
    n3 = Node(d3)

    model2 = BayesianNetwork()
    model2.add_nodes(n1, n2, n3)
    model2.add_edge(n1, n2)
    model2.add_edge(n2, n3)
    model2.bake()
    model2.fit(data_generator)

    logp1 = model1.log_probability(X)
    logp2 = model2.log_probability(X)

    assert_array_almost_equal(logp1, logp2)

@with_setup(setup_random_mixed, teardown)
def test_io_from_samples():
    model1 = BayesianNetwork.from_samples(X, weights=weights)
    model2 = BayesianNetwork.from_samples(data_generator)

    logp1 = model1.log_probability(X)
    logp2 = model2.log_probability(X)

    assert_array_almost_equal(logp1, logp2)

@with_setup(setup_random_mixed, teardown)
def test_io_from_structure():
    structure = ((2,), (0, 2), ())

    model1 = BayesianNetwork.from_structure(X=X, weights=weights,
        structure=structure)
    model2 = BayesianNetwork.from_structure(X=data_generator,
        structure=structure)

    logp1 = model1.log_probability(X)
    logp2 = model2.log_probability(X)

    assert_array_almost_equal(logp1, logp2)
